#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

import logging
import dataclasses
import jwt.exceptions
from functools import wraps
from typing import Tuple, NewType, cast

from flask import Request, current_app
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import provider
    from flask_injector import request as request_scope
    from ..services.view_exceptions import (
        Unauthorized, Forbidden, UnprocessableEntity
    )

except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def provider(function):  # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper

    def request_scope(function):  # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)

        return wrapper

    # dont need to patch exception, wont be used by SDK


from ..providers.base_provider import BaseModule


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Config:
    key: str
    audience: str
    algorithms: Tuple[str] = ('HS256',)
    cookie_name: str = 'oidc_id_token'
    auth_header_name: str = 'Authorization'
    bearer_prefix: str = 'Bearer '
    # TODO scott: After new pyjwt release, if the JWTs from Catalogue start failing because of these options then
    #  should they be set to False?
    verify_signature: bool = True
    verify_aud: bool = True
    verify_exp: bool = True

    aws_prefix: str = 'AWS'


JWT = NewType('JWT', dict)
JWTEncoded = NewType('JWTEncoded', str)


class OIDCDependencyProvider(BaseModule):
    config_class = Config

    @classmethod
    def _get_aws_jwt(cls, header_jwt):
        """
        Pulls the AWS Access key Id from a boto request.

        This is completely redundant if going via middleware (the apps/integration currently dont though)
        """
        if header_jwt.startswith('AWS4-HMAC-SHA256 '):
            _, _, data = header_jwt.partition(' ')
            for aws_credential in data.split(','):
                type_, _, cred_data = aws_credential.strip().partition('=')

                if type_ == 'Credential':
                    access_key_id, _, _ = cred_data.partition('/')
                    return access_key_id

        if header_jwt.startswith('AWS '):
            processed_header = header_jwt[len('AWS '):]
            access_key_id, _, _ = processed_header.strip().partition(':')
            return access_key_id

        logger.warning(
            'Malformed AWS Request',
            extra={
                'header': header_jwt
            }
        )

        raise UnprocessableEntity(
            'Malformed AWS Request.'
        )

    @request_scope
    @provider
    def user_info(self, jwt_encoded: JWTEncoded) -> JWT:
        """
        Verifies cookie and yields JWT

        ----

        I want to make clear we're doing a few things wrong with
        authentication.

        Retrieving the OIDC token is the job of the old application.
        Towards the end of the authentication flow a JWT cookie is set.

        The way the JWT cookie is set is completely wrong.

        Flask-OIDC (used in the old app) ignores the signature of the JWT
        issued by Hydra, then creates it's own SHA256 hash of of the header
        and payload, literally doing the following.

        >>> generated_signature = hmac.new(
        >>>     flask_secret_key,
        >>>     msg=header_and_payload,
        >>>     digestmod=hashlib.sha256
        >>> ).hexdigest()
        >>> assert generated_signature == signature
        """
        try:
            options = {
                "verify_signature": self.config.verify_signature,
                "verify_aud": self.config.verify_aud,
                "verify_exp": self.config.verify_exp,
            }
            payload = jwt.decode(
                jwt_encoded,
                self.config.key,
                audience=self.config.audience,
                algorithms=list(self.config.algorithms),
                options=options
            )
            pl = JWT(payload)
            return pl  # TODO refactor to call new class here
        except jwt.exceptions.InvalidSignatureError as e:
            # this is a configuration issue - check that JWT_SECRET_KEY/OIDC_SECRET_KEY is correct for/and correct env
            raise Unauthorized(
                'User not authenticated. Invalid Token: ' + str(e) + '.'
            )
        except jwt.exceptions.InvalidTokenError as e:
            raise Unauthorized(
                'User not authenticated. Invalid Token: ' + str(e) + '.'
            )
        except TypeError:
            logger.exception(
                f"JWT type error",
                extra={
                    'jwt_encoded': jwt_encoded,
                    'jwt_encoded type': type(jwt_encoded),
                    'config': self.config,
                },
            )
            raise

    @request_scope
    @provider
    def jwt(self, request: Request) -> JWTEncoded:

        cookie_jwt = request.cookies.get(self.config.cookie_name)
        header_jwt = request.headers.get(self.config.auth_header_name)

        # case: both
        if cookie_jwt and header_jwt:
            raise UnprocessableEntity(
                'User cannot be authenticated. '
                'Only one token should be '
                'provided either in the '
                'Authorization Header or in '
                'the Cookie.'
            )

        # case: neither
        if not cookie_jwt and not header_jwt:
            raise Unauthorized(
                'User not authenticated. Token '
                'has to be provided either '
                'in the Authorization Header .'
                # not mentioning cookie anymore here, as this is for QA/Catalogue only
                # we don't want external users knowing about it.
            )

        # case: either-or
        if cookie_jwt:
            return JWTEncoded(cookie_jwt)
        else:
            header_jwt = cast(str, header_jwt)

            if header_jwt.startswith(self.config.bearer_prefix):
                return JWTEncoded(header_jwt.split(' ')[-1])
            if header_jwt.startswith(self.config.aws_prefix):
                '''        
                    This is completely redundant if going via middleware (the apps/integration currently dont though).
                    boto makes direct requests here presently (s3proxy), so needs to be able to exchange directly.
                    
                    TODO will be to make everything travel via the middleware, at which point any AWS exchange should be 
                    removed (29/NOV/2021)
                '''
                return JWTEncoded(self._get_aws_jwt(header_jwt))
            else:
                raise UnprocessableEntity(
                    'User cannot be authenticated. '
                    'Authorization Header type not '
                    'provided.'
                )


def public(view):
    """
    Marks view as public. Ignores authentication errors.
    """
    @wraps(view)
    def wrapper(*args, **kwargs):
        try:
            current_app.injector.get(JWT)
        except (Unauthorized, Forbidden):
            pass

        return view(*args, **kwargs)

    return wrapper


def private(view):
    """
    Marks view as private. Attempts to get a token from
    the request. If it fails the OIDC provider will tell
    us why.
    """

    # This will modifier the swag from result to include security data
    if hasattr(view, 'specs_dict'):
        view.specs_dict['security'] = [{'bearer_auth': []}]

    @wraps(view)
    def wrapper(*args, **kwargs):
        # raises error if it fails. Caches user object
        current_app.injector.get(JWT)

        return view(*args, **kwargs)

    return wrapper



import base64
import datetime
import json
from typing import Optional
from unittest import mock
from unittest.mock import MagicMock, call
import jwt
import pytest
import requests
from flask import Response, request, Blueprint

from ..authentication.middleware import HEADER_MODIFIER, RESPONSE_MODIFIER, ServiceMiddleware
from ..authentication import auth
from ..providers.oidc_provider import public
from ..tests.common import disconnected_client, mock_middleware_server
_ = disconnected_client  # stop import being marked as unused (fixture is used everywhere below)


class TestMiddleware:

    def test_wsgi_middleware_calls_expected_route(self, disconnected_client):
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")

        response = disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}"})
        assert response.status_code == 200
        assert response.data == b'HELLO, WORLD'

        response2: Response = disconnected_client.get("/direct_test", headers={"Authorization": f"Bearer {jw}"})
        assert response2.status_code == 403
        assert response2.data == b'BE GONE, DRAGON'

    def test_wsgi_middleware_calls_injected_header_modifier(self, disconnected_client):
        # check that once we attach them, they are called

        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        header_modifier = MagicMock()
        header_modifier.side_effect = lambda jwt, env: env
        disconnected_client.application.injector.binder.bind(
            HEADER_MODIFIER,
            to=lambda: header_modifier,
            scope=request,
        )
        disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}", "Another-Header": "findme"})
        header_modifier.assert_called()
        # argument list should be position[0] = jwt, position[1] = original headers (+ http/wsgi)
        header_modifier.assert_has_calls([
           call(jw, mock.ANY)
        ])
        # lets inspect those headers, make sure we have the one we passed
        recovered_headers_from_call_argument = header_modifier.call_args_list[0].args[1]
        assert "HTTP_ANOTHER_HEADER" in recovered_headers_from_call_argument
        # and that we popped the ones we didnt want already in middleware
        assert "HTTP_AUTHORIZATION" not in recovered_headers_from_call_argument
        assert "HTTP_HOST" not in recovered_headers_from_call_argument

    def test_wsgi_middleware_calls_injected_response_modifier(self, disconnected_client):
        # check that once we attach them, they are called

        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        ground_zero = disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}"})

        response_modifier = MagicMock()
        response_modifier.side_effect = lambda response: ("goodbye world", 403, {})
        disconnected_client.application.injector.binder.bind(
            RESPONSE_MODIFIER,
            to=lambda: response_modifier,
        )
        result = disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}"})

        class ResponseAny():
            def __eq__(self, other):
                return type(other) == requests.models.Response

        response_modifier.assert_called()
        response_modifier.assert_has_calls([
            call(ResponseAny())
        ])

        # check our responsemodifier has overwritten the response vis-a-vis calling without
        assert result.data == b"goodbye world"
        assert result.status_code == 403
        assert ground_zero.status_code != result.status_code

    def test_wsgi_middleware_injectors_called_before_and_after_route(self):
        # timing - we can check that within the route, the headers have been added/changed prior to route
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        header_mod_called = None
        view_called = None
        response_mod_called = None

        def timing_header(jwt, env):
            nonlocal header_mod_called
            header_mod_called = datetime.datetime.now()
            return env

        def timing_response(response):
            nonlocal response_mod_called
            # assert we've not changed the response yet
            response_mod_called = datetime.datetime.now()
            return "goodbye world", 403, {}

        def timing_app():
            views = Blueprint('middleware', __name__)
            HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']

            @views.route(
                '/',
                methods=HTTP_METHODS
            )
            def test_route(optional_path: Optional[str] = None):
                nonlocal view_called
                view_called = datetime.datetime.now()

                resp = MagicMock()
                resp.status_code = 200
                resp.headers = {}
                resp.content = "HELLO, WORLD"
                return resp.content, resp.status_code, resp.headers.items()

            timing_app = mock_middleware_server(views)
            with timing_app.test_client() as client:
                return client

        disconnected_client = timing_app()

        header_modifier = MagicMock()
        header_modifier.side_effect = timing_header
        disconnected_client.application.injector.binder.bind(
            HEADER_MODIFIER,
            to=lambda: header_modifier,
            scope=request,
        )
        response_modifier = MagicMock()
        response_modifier.side_effect = timing_response
        disconnected_client.application.injector.binder.bind(
            RESPONSE_MODIFIER,
            to=lambda: response_modifier,
        )

        disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}"})
        assert header_mod_called < view_called < response_mod_called

    def test_wsgi_modifies_headers_correctly_before_route(self, disconnected_client):
        # correctness - that the headers required are changed

        # here we will devise any incoming auth header should be intercepted and resolved by the time
        # we are in the view (be already transformed to custom auth header format)

        # example - our real endpoint only takes Authorisation: Bearer <TOKEN>, but our middleware we enter User:Pass
        #   we want to transform the Authorisation header into that form by exchanging the User:Pass -> Token prior
        #   to it arriving in the route which sends it onto the endpoint

        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        intercepted_headers = None

        # underscores are flipped to dashes in the wsgi container, HTTP_ prefix is dropped.
        # so additional headers may not use "_"
        # i.e. HTTP_X_SOURCE = X-Source
        additional_headers = {
            "X-Additional-Header1": "tag",
            "X-Source": "middleware",
            "X-Source-User": "<JWT>",
            "Cache-Control": "no-cache",
            "Connection": 'keep-alive',
            "Content-Type": "application/json"
        }

        def mod_header(jwt, env):
            new_headers = env.copy()

            # some kind of exchange has already taken place to get us a JWT from our creds in the params
            # so we place that into the bearer

            dd = {"HTTP_AUTHORIZATION": f"Bearer {jwt}"}

            # must be prefixed HTTP_ else won't end up in the middleware headers provided to the route
            # this way we also preserve the dash "-" which would otherwise be converted to underscore "_"
            # use 'HACK' of : dict([x for x in request.headers]) to read these back out without werkzeug throwing a fit
            for k, v in additional_headers.items():
                if v == "<JWT>":
                    dd["HTTP_"+k] = jwt
                else:
                    dd["HTTP_" + k] = v

            new_headers.update(dd)

            return new_headers

        def inspect_header_app():
            views = Blueprint('middleware', __name__)
            HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']

            @views.route(
                '/',
                methods=HTTP_METHODS
            )
            def test_route(optional_path: Optional[str] = None):

                # inspect the headers here to check they are already transformed as we would need to
                # make a request from here to an EP.
                nonlocal intercepted_headers
                intercepted_headers = request.headers

                def session():
                    return requests.Session()

                resp = MagicMock()
                resp.status_code = 200
                resp.headers = {}
                resp.content = "HELLO, WORLD"

                session()

                return resp.content, resp.status_code, resp.headers.items()

            inspect_app = mock_middleware_server(views)
            with inspect_app.test_client() as client:
                return client

        # basic auth is going to go down this exchange to get the JWT passed to our header modifier func
        auth._sam_app_flow = MagicMock()
        auth._sam_app_flow.return_value = jw

        # pat auth is going to go down this exchange to get the JWT passed to our header modifier func
        auth._sam_pat_flow = MagicMock()
        auth._sam_pat_flow.return_value = jw

        # jwt auth is going to return the JWT (the Identity function) so requires no mock
        _ = None

        disconnected_client = inspect_header_app()
        disconnected_client.application.injector.binder.bind(
            HEADER_MODIFIER,
            to=lambda: mod_header,
            scope=request,
        )

        # the LDAP/Presto case for PAT or APP cases
        user_pass = base64.urlsafe_b64encode(b"<app_user>:<app_pasw>").decode("ascii")
        disconnected_client.get("/", headers={"Authorization": f"Basic {user_pass}"})
        assert intercepted_headers["Authorization"] == f"Bearer {jw}"

        # the SDK case for Presto, passing JWT, ends up as the Identity function
        disconnected_client.get("/", headers={"Authorization": f"Bearer {jw}"})
        assert intercepted_headers["Authorization"] == f"Bearer {jw}"

        # the PAT or APP case, passing delimited user/pass secrets in the signature via Spark/Boto/tools (S3Proxy)
        sig = ("Credential=<app_user>\\<app_pasw>/<date>/<region>/<aws_service>/aws4_request, "
               "SignedHeaders=amz-sdk-invocation-id;amz-sdk-retry;content-type;host;user-agent;x-amz-content-sha256;x-amz-date, "
               "Signature=<calc_signature>")
        disconnected_client.get("/", headers={"Authorization": f"AWS4-HMAC-SHA256 {sig}"})
        assert intercepted_headers["Authorization"] == f"Bearer {jw}"

        # the SDK case for S3Proxy, passing JWT in the signature via boto (S3Proxy)
        sig2 = (f"Credential={jw}/<date>/<region>/<aws_service>/aws4_request, "
               "SignedHeaders=amz-sdk-invocation-id;amz-sdk-retry;content-type;host;user-agent;x-amz-content-sha256;x-amz-date, "
               "Signature=<calc_signature>")
        disconnected_client.get("/", headers={"Authorization": f"AWS4-HMAC-SHA256 {sig2}"})
        assert intercepted_headers["Authorization"] == f"Bearer {jw}"

        # we also add some additional headers, that we wanted to pass along to the service for each request.
        # lets check those were added in mod_header (again, recall they are not going to be prefixed HTTP_ inside
        # of the wsgi container), i.e. "HTTP_something" in fn: mod_header == "something" in flask view

        dict_of_modded_headers = dict([x for x in intercepted_headers])
        for k, v in additional_headers.items():
            assert k in dict_of_modded_headers
            assert dict_of_modded_headers[k] == (v if not v == "<JWT>" else jw)

    @pytest.mark.xfail
    def test_all_headers_data_sent_to_proxied_service(self, disconnected_client):
        # TBC - there isnt really a test for this, since its the app's responsibility in the view to decide
        # this is just to notate how to use those pesky headers that we hacked

        '''
            headers = dict([x for x in request.headers])  # headers back to dict to allow header[X] use in view
            resp = getattr(requests, request.method.lower())(upstream_service, headers=headers, data=request.data)
        '''

    def test_wsgi_modifies_response_correctly_after_route(self, disconnected_client):
        # correctness - that the content/headers are modified how we want

        # here we will devise any outgoing response should be intercepted and
        # transformed before sending to client. Lets imagine our service gives a json response with
        # `follow` urls in the json for the client to make the consequent request.

        # We need to modify those `follow` urls to redirect the consequent request via the middleware rather than
        # directly back towards the service (which may not be visible to them, nor permit their auth).

        # The client should then receive this modified response instead.
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")

        def mod_response(response):
            content = response.content
            a = json.loads(content)
            a["nextUri"] = a["nextUri"].replace("my-service", "my-middleware")
            a["inspectUri"] = a["inspectUri"].replace("my-service", "my-middleware")
            return json.dumps(a), response.status_code, response.headers.items()

        def response_from_app():
            views = Blueprint('middleware', __name__)
            HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']

            @views.route(
                '/',
                methods=HTTP_METHODS
            )
            def test_route(optional_path: Optional[str] = None):

                resp = MagicMock()
                resp.status_code = 200
                resp.headers = {}

                json_body = {
                    "nextUri" : "http://my-service/1234",
                    "inspectUri": "http://my-service/inspect/1234?slug=vi54nvg65"
                }

                resp.content = json.dumps(json_body)

                return resp.content, resp.status_code, resp.headers.items()

            mock_response_app = mock_middleware_server(views)
            with mock_response_app.test_client() as client:
                return client

        # skip auth (patch before app)
        ServiceMiddleware._resolve = MagicMock()
        ServiceMiddleware._resolve.return_value = jw

        disconnected_client = response_from_app()
        disconnected_client.application.injector.binder.bind(
            RESPONSE_MODIFIER,
            to=lambda: mod_response,
            scope=request,
        )

        resp = disconnected_client.get("/", headers={"Authorization": f"Basic anything"})
        _json = json.loads(resp.data)
        assert _json["nextUri"] == "http://my-middleware/1234"
        assert _json["inspectUri"] == "http://my-middleware/inspect/1234?slug=vi54nvg65"

    @pytest.mark.xfail # TODO
    def test_wsgi_respond_from_bad_reply_from_proxied_service(self, disconnected_client):
        # correctness - that say a 403/500 from the service is sent back through correctly
        pass

    @pytest.mark.xfail # TODO
    def test_public_route_allowed(self, disconnected_client):
        # correctness - any public routes will be trapped behind auth of the middleware, unless handled
        def public_app():
            views = Blueprint('middleware', __name__)
            HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']

            @views.route(
                '/',
                methods=HTTP_METHODS
            )
            @public
            def test_route(optional_path: Optional[str] = None):

                resp = MagicMock()
                resp.status_code = 200
                resp.headers = {}

                json_body = {
                    "nextUri" : "http://my-service/1234",
                    "inspectUri": "http://my-service/inspect/1234?slug=vi54nvg65"
                }

                resp.content = json.dumps(json_body)

                return resp.content, resp.status_code, resp.headers.items()

            mock_response_app = mock_middleware_server(views)
            with mock_response_app.test_client() as client:
                return client

        disconnected_client = public_app()
        resp = disconnected_client.get("/")
        assert resp.status_code == 200


import base64
import os
import uuid
from unittest.mock import MagicMock, patch
import pytest
import jwt

from requests import Response

from ..context.environments import _Environment
from ..authentication.middleware import ServiceMiddleware
from ..authentication.auth import _sam_app_flow, _sam_pat_flow, _pre_request_get_token, decode_token, Unauthorized
from ..authentication import auth

from ..tests.common import disconnected_client
_ = disconnected_client  # stop import being marked as unused (fixture is used everywhere below)


class TestAuth:

    def test_no_auth(self, disconnected_client):
        a = disconnected_client.get("/test")
        assert a.status_code == 401
        assert a.data.decode() == 'No Authorization header provided'

    def test_unrecognized_auth(self, disconnected_client):
        with pytest.raises(Unauthorized):
            disconnected_client.get("/test", headers={"Authorization": "Random XXXXX"})

    def test_unrecognized_jwt_decode(self, disconnected_client):
        with pytest.raises(jwt.DecodeError):
            disconnected_client.get("/test", headers={"Authorization": "Bearer JWT"})

    def test_recognized_jwt_hits_backend(self, disconnected_client):
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        resp = disconnected_client.get("/test", headers={"Authorization": f"Bearer {jw}"})
        assert resp.status_code == 200
        assert resp.data.decode() == "HELLO, WORLD"

    def test_resolver_is_called(self, disconnected_client):
        with patch.object(ServiceMiddleware, '_resolve', return_value=MagicMock()) as mock_method:
            jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
            disconnected_client.get("/test", headers={"Authorization": f"Bearer {jw}"})
            mock_method.assert_called()

    def test_resolver_calls_correct_method_for_auth_type(self, disconnected_client):

        # make all the caller methods magicmocks
        jw = jwt.encode(payload={"<jwt_user>": "payload"}, key="secret", algorithm="HS256")

        with patch.object(ServiceMiddleware, '_type_bearer', return_value=(jw, "NOP", "JWT")) as _bearer,\
             patch.object(ServiceMiddleware, '_type_ldap', return_value=("app_user", "app_pasw", None)) as _ldap,\
             patch.object(ServiceMiddleware, '_type_s3_rest', return_value=("app_user", "app_pass", None)) as _rest,\
             patch.object(ServiceMiddleware, '_type_v4_signature', return_value=("app_user", "app_pass", None)) as _sig,\
             patch.object(ServiceMiddleware, '_credentials_to_token', return_value=MagicMock()
        ):

            disconnected_client.get("/test", headers={"Authorization": f"Bearer {jw}"})
            # check only this one was called
            assert _bearer.call_count == 1
            assert _ldap.call_count == 0
            assert _sig.call_count == 0
            assert _rest.call_count == 0
            # check we called it how we expected
            _bearer.assert_called_with("\\", jw)
            _bearer.reset_mock()

            ldap = base64.urlsafe_b64encode(b"<app_user>:<app_pasw>")
            disconnected_client.get("/test", headers={"Authorization": f"Basic {ldap}"})
            # check only this one was called
            assert _bearer.call_count == 0
            assert _ldap.call_count == 1
            assert _sig.call_count == 0
            assert _rest.call_count == 0
            _ldap.assert_called_with("\\", str(ldap))
            _ldap.reset_mock()

            # https://docs.aws.amazon.com/AmazonS3/latest/userguide/RESTAuthentication.html
            rest ="<app_user\\app_pasw>:<signature>"
            disconnected_client.get("/test", headers={"Authorization": f"AWS {rest}"})
            _rest.assert_called_with("\\", rest)

            # https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
            sig = ("Credential=<app_user\\app_pasw>/<date>/<region>/<aws_service>/aws4_request, "
                "SignedHeaders=amz-sdk-invocation-id;amz-sdk-retry;content-type;host;user-agent;x-amz-content-sha256;x-amz-date, "
                "Signature=<calc_signature>")
            disconnected_client.get("/test", headers={"Authorization": f"AWS4-HMAC-SHA256 {sig}"})
            _sig.assert_called_with("\\", sig)

    def test_split_jwt_credentials(self):
        colon_case_present = ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter=":", field_aws_access_key_id="user:password")
        assert colon_case_present == ("user", "password", None)

        backslash_case_present = ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter="\\", field_aws_access_key_id="user\\password")
        assert backslash_case_present == ("user", "password", None)

        JWT = "jwt005"
        divider_missing = ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter="\\", field_aws_access_key_id=JWT)
        assert divider_missing == (JWT, "NOP", "JWT")

        # we have confirmed the relevant divider cannot be present in: the user+header auth (CC - [A-Za-z0-9], or PAT - [UUID])
        # we have confirmed the relevant divider "\\" is not/cannot be present in the SAM PAT password (thus the choice)
        divider_in_password = ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter="\\", field_aws_access_key_id="user\\pass\\w\\ord")
        assert divider_in_password == ("user", "pass\\w\\ord", None)

    def test_type_ldap(self):
        ldap = base64.urlsafe_b64encode(b"<app_user>:<app_pasw>")
        creds = ServiceMiddleware._type_ldap("\\", ldap)
        assert creds == ("<app_user>", "<app_pasw>", None)

    def test_type_bearer(self):
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        creds = ServiceMiddleware._type_bearer("\\", jw)
        assert creds == (jw, "NOP", "JWT")

    def test_type_s3_rest(self):
        rest ="<app_user>\\<app_pasw>:<signature>"
        creds = ServiceMiddleware._type_s3_rest("\\", rest)
        assert creds == ("<app_user>", "<app_pasw>", None)

    def test_type_v4_signature(self):
        sig = ("Credential=<app_user>\\<app_pasw>/<date>/<region>/<aws_service>/aws4_request, "
               "SignedHeaders=amz-sdk-invocation-id;amz-sdk-retry;content-type;host;user-agent;x-amz-content-sha256;x-amz-date, "
               "Signature=<calc_signature>")
        creds = ServiceMiddleware._type_v4_signature("\\", sig)
        assert creds == ("<app_user>", "<app_pasw>", None)

    def test_sam_app_flow(self):
        # we aren't going overboard with testing on this method, since it will move to catalogue
        env = _Environment("https://catalogue-dev.udpmarkit.net")
        auth.external_session = MagicMock()

        sam_response = Response()
        sam_response._content = b'{"access_token": "<JWT>", "id_token": "<JWT2>"}'
        sam_response.headers.update({})
        sam_response.status_code = 200

        catalogue_response = Response()
        catalogue_response._content = b'{"access_token": "<JWT3>"}'
        catalogue_response.headers.update({})
        catalogue_response.status_code = 200

        auth.external_session.return_value.post.side_effect = [
            sam_response, catalogue_response
        ]
        jwt = _sam_app_flow(env, "user", "pasw")
        assert jwt == "<JWT3>"

    def test_sam_pat_flow(self):
        # we aren't going overboard with testing on this method, since it will move to catalogue
        env = _Environment("https://catalogue-dev.udpmarkit.net")

        auth.external_session = MagicMock()

        sam_response = Response()
        sam_response._content = b'{"access_token": "<JWT>", "id_token": "<JWT2>"}'
        sam_response.headers.update({})
        sam_response.status_code = 200

        catalogue_response = Response()
        catalogue_response._content = b'{"access_token": "<JWT3>"}'
        catalogue_response.headers.update({})
        catalogue_response.status_code = 200

        auth.external_session.return_value.post.side_effect = [
            sam_response, catalogue_response
        ]
        os.environ["SAM_CLIENT_ID_FOR_PAT_EXCHANGE"] = "PAT_ID"
        os.environ["SAM_CLIENT_SECRET_FOR_PAT_EXCHANGE"] = "PAT_SECRET"

        jwt = _sam_pat_flow(env, "user", "pasw")
        assert jwt == "<JWT3>"

    def test_decode_token(self):
        good_jwt = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        assert decode_token(good_jwt) == {"some": "payload"}

        bad_jwt = "not_a_jwt"
        with pytest.raises(Exception) as execinfo:
            decode_token(bad_jwt)

    def test_pre_exchange_jwt_flow(self):
        env = _Environment("https://catalogue-dev.udpmarkit.net")

        with patch.object(auth, '_sam_pat_flow', return_value=MagicMock()) as patflow:
            with patch.object(auth, '_sam_app_flow', return_value=MagicMock()) as appflow:
                with patch.object(auth, 'decode_token', return_value=MagicMock()) as decodetoken:
                    # note - this test is precedence ordered, do not change order
                    # (must match same order flow as _pre_request_get_token): PAT->JWT->APP

                    # hinted PAT case, and provided PAT
                    r = _pre_request_get_token(env, str(uuid.uuid4()), "<pass>", "PAT")
                    patflow.assert_called()

                    # hinted PAT case, and not provided PAT
                    with pytest.raises(ValueError) as execinfo:
                        _pre_request_get_token(env, "APP_USER", "<pass>", "PAT")

                    # hinted JWT case and provided JWT
                    jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
                    _pre_request_get_token(env, str(jw), "<pass>", "JWT")
                    decodetoken.assert_called()

                # hinted JWT case and not provided JWT
                with pytest.raises(jwt.DecodeError) as execinfo:
                    _pre_request_get_token(env, str(uuid.uuid4()), "<pass>", "JWT")

                # hinted APP case, and provided APP
                _pre_request_get_token(env, "APP_USER", "<pass>", "APP")
                appflow.assert_called()

            # hinted APP case, and not provided APP
            # we now need to patch requests post to be a 403, as its actually gonna go to the EP
            resp = MagicMock()
            resp.post.return_value = MagicMock()
            resp.post.return_value.status_code = 403
            with patch.object(auth, 'external_session', return_value=resp) as response:
                with pytest.raises(Unauthorized) as execinfo:
                    _pre_request_get_token(env, str(uuid.uuid4()), "<pass>", "APP")

    def test_pre_exchange_jwt_flow_unhinted(self):
        env = _Environment("https://catalogue-dev.udpmarkit.net")
        auth._sam_pat_flow = MagicMock()
        auth._sam_app_flow = MagicMock()
        auth._decode_token = decode_token
        auth.sam_app_flow = _sam_app_flow
        auth.decode_token = MagicMock()

        # not provided username/pass
        with pytest.raises(Unauthorized) as execinfo:
            _pre_request_get_token(env, None, None, None)

        # unhinted PAT case, and provided PAT
        _pre_request_get_token(env, str(uuid.uuid4()), "<pass>", None)
        auth._sam_pat_flow.assert_called()
        auth._sam_pat_flow.reset_mock()

        # unhinted JWT case, and provided JWT
        jw = jwt.encode(payload={"some": "payload"}, key="secret", algorithm="HS256")
        _pre_request_get_token(env, str(jw), "<pass>", None)
        auth.decode_token.assert_called()
        auth.decode_token.reset_mock()

        # unhinted APP case, and provided APP
        auth.decode_token = auth._decode_token
        _pre_request_get_token(env, "APP_USER", "<pass>", None)
        auth._sam_app_flow.assert_called()
        auth._sam_app_flow.reset_mock()

        # unhinted case, and not valid
        auth._sam_app_flow = auth.sam_app_flow
        with pytest.raises(Unauthorized) as execinfo:
            _pre_request_get_token(env, "<user>", "<pass>", None)


import logging
import os
from typing import Optional

import requests

logger = logging.getLogger(__name__)


class ReportHandler:

    def __init__(self):
        self.report_host = os.environ.get("REPORTING_URL", "None")

    def report(
        self,
        statement: str,
        status: bool = True,
        dataset_shortcode: Optional[str] = None,
        modified: Optional[bool] = None,
        stage: Optional[str] = None,
        note1: Optional[str] = "",
        note2: Optional[str] = ""
    ):
        # only report 'stage' based things for a dataset or the process as a whole, not logs.
        # we don't want to know you're getting session details at a high level
        # e.g. beginning Unit Z, ending Unit Y
        status = "success" if status else "fail"
        level = "dataset" if dataset_shortcode is not None else "collector"
        modified = "yes" if modified else "no"
        if stage not in ["extract", "transform", "load", "validation"]:
            logger.warning('Stage is not set to extract / transform / load / validation')

        data = {
            "statement": statement,
            "modified": "no"
        }

        if dataset_shortcode:
            data["dataset"] = dataset_shortcode
            data["modified"] = modified
            data["note1"] = note1
            data["note2"] = note2

        try:
            if self.report_host != 'None':
                requests.post(
                    f"{self.report_host}/report/{stage}/{level}/{status}",
                    json=data,
                    timeout=0.5
                )
        except Exception as e:
            # back off and retry
            logger.debug("[crawler] Cannot report")




from dli.models.dictionary_model import DictionaryModel

from platform_services_lib.lib.context.urls import dataset_urls


class Dictionary:

    def register_dictionary(
        self,
        dataset_id,
        version,
        valid_as_of,
        fields,
        **kwargs
    ):
        """
        Registers dictionary metadata for a dataset.

        :param str dataset_id: Id of the dataset for the dictionary.
        :param str version: A user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: The date as of which the dictionary is active.
                               Expected format is YYYY-MM-DD. Must be unique.
        :param list[dict] fields: Non empty list of `Field` as described below.
        :param list[dict] partitions: Optional. Non empty list of `Partition` as described below.
        :param str description: Optional. Description for the dictionary.

        :returns: The registered dictionary

        Types
        =====

        Dictionary Field:

        .. code-block:: python

            {
                name	        string  - required
                type	        string  - required
                nullable        boolean - required
                metadata	dictionary
                description	string
                sample_value	string
                short_name	string
                is_derived	boolean
                validation	string
                comment	        string
            }

        Partition:

        .. code-block:: python

            {
                name: string,
                type: string
            }

        - **Sample**

        .. code-block:: python

                my_dictionary_fields = [
                            {
                                'name': 'field_1',
                                'type': 'String',
                                'nullable': False
                            },
                            {
                                'name': 'field_2',
                                'type': 'Double',
                                'nullable': False
                            },
                            {
                                'name': 'field_3',
                                'type': 'Int',
                                'nullable': True,
                                'metadata': {
                                    'some_key': 'some_value'
                                }
                            },
                        ]
                my_dictionary_partitions = [
                    {
                        'name': 'field_1',
                        'type': 'String'
                    }
                ]

                my_dictionary = client.register_dictionary(
                    "my-dataset-id",
                    version='1a',
                    valid_as_of='2018-10-31',
                    fields=my_dictionary_fields,
                    partitions=my_dictionary_partitions,
                    description="My dictionary description"
                )
        """
        payload = {
            'dataset_id': dataset_id,
            'version': version,
            'valid_as_of': valid_as_of,
            'fields': fields,
        }

        payload.update(**kwargs)

        response = self.session.post(
            dataset_urls.dictionary_index, json={'data': {'attributes': payload}}
        )

        return DictionaryModel(response.json()['data'], client=self)

    def get_dictionary(self, dataset_id, version=None):
        """
        Looks up dictionary for a dataset by version. In case version is not specified this will fetch dictionary version having the latest valid_as_of date.
        Throws exception if no dictionary is registered for the dataset.

        :param str dataset_id: The id of the dataset under which the dictionary is registered.
        :param str version: Optional. The version of the dictionary.

        :returns: The dictionary.

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                # Fetch dictionary version '1a'
                dictionary = client.get_dictionary('my_dataset_id', version='1a')

                # Fetch the dictionary with the latest valid_as_of date
                latest_dictionary = client.get_dictionary('my_dataset_id')

        """

        if version:
            schema = self.session.get(
                dataset_urls.v2_schema_instance_version.format(
                    id=dataset_id,
                    version=version
                )
            )
        else:
            schema = self.session.get(
                dataset_urls.dictionary_by_dataset_lastest.format(
                    id=dataset_id
                )
            )
        json = schema.json()['data']
        return DictionaryModel(json, client=self)

    def get_dictionaries(self, dataset_id, count=100):
        """
        Returns a list of all dictionaries registered under the dataset. The list is sorted by dictionary valid_as_of date in descending order.

        :param str dataset_id: The id of the dataset.
        :param int count: Optional count of dictionaries to be returned. Defaults to 100.

        :returns: List of all dictionaries registered under the dataset sorted by valid_as_of date in descending order.

        - **Sample**

        .. code-block:: python

                my_dictionaries = client.get_dictionaries("my_dataset_id", count=10)
        """
        response = self.session.get(
            dataset_urls.dictionary_by_dataset.format(id=dataset_id)
        )

        return [
            DictionaryModel(d, client=self) for d in
            response.json()['data']
        ][:count]  # todo - will kill this silly count stuff ASAP

    def delete_dictionary(self, dataset_id, version):
        """

        Marks a dictionary version for a dataset as deleted.

        :param str dataset_id: The id of the dataset under which the dictionary is registered.
        :param str version: The version of the dictionary.

        :returns: None

        - **Sample**

        .. code-block:: python

                # Delete dictionary version '1a'

                client = dli.connect()
                client.delete_dictionary(dataset_id='my_dataset_id', version='1a')

        """
        schema = self.get_dictionary(dataset_id, version)
        self.session.delete(
            dataset_urls.dictionary_instance.format(id=schema.id)
        )

    def edit_dictionary(
        self,
        dataset_id,
        version,
        new_version=None,
        **kwargs
    ):
        """
        Updates dictionary metadata for a dataset.
        If a field is passed as ``None`` then the field will not be updated.

        :param str dataset_id: Id of the dataset for the dictionary_instance.
        :param str version: Version of the dictionary being updated.
        :param str new_version: Optional. New version if to be updated. This is a user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: Optional. The date as of which the dictionary is active.
                               Expected format is YYYY-MM-DD.
        :param list[dict] fields: Optional. If provided, a non empty list of `Field` as described below.
        :param list[dict] partitions: Optional. If provided, a non empty list of `Partition` as described below.
        :param str description: Optional. Description for the dictionary.

        .. code-block:: python

                # Field:
                {
                    "name": "field_a", 			# name of the column.
                    "nullable": True,  			# defaulted to True - A boolean indicating whether the field is nullable or not.
                    "metadata": None			# optional dictionary with metadata for this column.
                }

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }

        :returns: The updated dictionary.
        :rtype: dli.models.dictionary_model.DictionaryModel

        - **Sample**

        .. code-block:: python

                # Updating description and valid_as_of date for my dictionary
                client = dli.connect()
                my_updated_schema = client.edit_dictionary(
                    "my-dataset-id",
                    '1a',
                    valid_as_of='2018-11-05',
                    description="My updated dictionary description"
                )
        """
        schema = self.get_dictionary(dataset_id, version)
        payload = dict(**kwargs)

        if new_version is not None:
            payload['version'] = new_version

        response = self.session.patch(
            dataset_urls.dictionary_instance.format(id=schema.id),
            json={'data': {'attributes': payload}}
        )

        return DictionaryModel(response.json()['data'], client=self)


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import multiprocessing
from typing import List, Optional
from concurrent.futures.thread import ThreadPoolExecutor
from urllib.parse import urljoin
import humps  # noqa: I900

from dli.models.file_model import FileModel

from platform_services_lib.lib.context.urls import consumption_urls
from platform_services_lib.lib.aspects.decorators import analytics_decorator, logging_decorator, log_public_functions_calls_using
from platform_services_lib.lib.services.dlc_attributes_dict import AttributesDict

THREADS = multiprocessing.cpu_count()


class InstanceModel(AttributesDict):

    def __init__(self, **kwargs):
        # Ignore the datafile's files
        kwargs.pop('files', [])
        super().__init__(**kwargs)

    def files(self) -> List[FileModel]:
        """
        :return: list of file models for files in the instance.
        """
        url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_manifest.format(
                id=self.datafile_id
            )
        )

        response = self._client.session.get(url)
        return [
            self._client._File(
                datafile_id=self.datafile_id,
                **d['attributes']
            )
            for d in response.json()['data']
        ]

    @classmethod
    def _from_v1_entity(cls, entity, page_size=None):
        properties = humps.decamelize(entity['properties'])
        return cls(**properties)

    def download(self, to: Optional[str]='./', flatten: Optional[bool]=False) -> List[str]:
        """
        Download the files for the instance, then return a list of the file
        paths that were written.

        :param str to: The path on the system, where the files
            should be saved. must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The flatten parameter default behaviour
            (flatten=False) will allow the user to specify that they would
            like to keep the underlying folder structure when writing the
            downloaded files to disk.
            When “flatten” is set to True the
            folder structure and download all data to the specified path in
            and retains only the original file name. This is not useful if the
            file names are identical for multiple files.

        :return: List of paths to the downloaded files.

        :example:

            Downloading without flatten:

            .. code-block:: python

                >>> dataset = client.datasets.get(dataset_short_code='ExampleDatasetShortCode', organisation_short_code='IHSMarkit')
                >>> dataset.instances.latest().download('./local/path/')
                [
                  'as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'as_of_date=2019-09-11/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]


        :example:

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:

            .. code-block:: python

                >>> dataset.instances.latest().download('./local/path/', flatten=True)
                [
                  'StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]


        """
        files = self.files()

        def _download(file_: FileModel):
            return file_.download(to=to, flatten=flatten)

        threads = min(THREADS, max(1, len(files)))
        with ThreadPoolExecutor(max_workers=threads) as executor:
            return list(executor.map(_download, files))

    def __str__(self):
        return f"\nINSTANCE {self.datafile_id} (size: {self.total_size/1048576 :.2f} Mb)"




log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['datafile_id']
)(InstanceModel)





#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import dataclasses
from urllib import parse
from os import environ
import requests
import logging
from flask import current_app
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import inject
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper
except:
    pass

from ..analytics_service.model import AnalyticsEvent
from ...providers.oidc_provider import JWT
from ...providers.ga_session_provider import GAClient
from ...services.cached_fetcher_service import CachedFetcherService


@dataclasses.dataclass
class Config:
    tracking_id: str = environ.get('GA_TID', '')
    ga_request_timeout: float = float(environ.get('GA_REQUEST_TIMEOUT', '60'))


class GoogleAnalyticsService:
    logger = logging.getLogger(__name__)

    __params_mapping = {
        'user_id': 'cd1',
        'organisation_id': 'cd2',
        'package_id': 'cd3',
        'dataset_id': 'cd4',
        'dictionary_id': 'cd5',
        'datafile_id': 'cd6',
        'name': 'cd7',
        'dataset_name': 'cd8',
        # This was done for the problem with PowerBi masking the first field. Do not remove.
        'user_id_again': 'cd9',
    }

    __ga_version_field = 'v'
    __ga_tracking_id_field = 'tid'
    __ga_type_field = 't'

    ga_unhandled_exception_message = \
        'Problem with sending analytics to Google. ' \
        'The user should be unaffected but developers should ' \
        'check we do not have a Google Analytics problem that ' \
        'is not related to the connection pool e.g. network ' \
        'connectivity.'

    @inject
    def __init__(
            self,
            ga_client: GAClient,
            config: Config,
            identity_service: CachedFetcherService,
    ):
        self._ga_client = ga_client
        self._config = config
        self.__identity_service = identity_service
        if not self._config.tracking_id:
            raise ValueError('GA Tracking ID has to be defined.')
        self._static_params = (
            (self.__ga_version_field, 1),
            (self.__ga_tracking_id_field, self._config.tracking_id),
            (self.__ga_type_field, 'event')
        )
        if not self._config.ga_request_timeout:
            raise ValueError('GA request timeout has to be defined.')

    def send_message(self, body):
        """
        :param body: url encoded payload to send to Google Analytics.
        :return: request
        """
        try:
            self._ga_client.post(
                '',
                data=body,
                timeout=self._config.ga_request_timeout
            )
        except ConnectionResetError:
            # https://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
            self.logger.warning('GA ConnectionResetError')
        except requests.exceptions.ConnectionError:
            self.logger.warning('GA ConnectionError connection aborted')
        except ConnectionAbortedError:
            self.logger.warning('GA ConnectionAbortedError')
        except Exception:
            # Log the exception, we should not raise analytics exceptions
            # back to the user, instead log them as exceptions for developers
            # to deal with.
            self.logger.exception(self.ga_unhandled_exception_message)

    def create_message_body(self, action_info: AnalyticsEvent):
        # Lookup action_info.user_id (which is an email address in Catalogue to get
        # the user_id so that we do not leak user email addresses.
        # Use the sub in the key so that it is a result for this
        # specific user.

        jwt = current_app.injector.get(JWT)

        me = self.__identity_service.dlc_me(jwt=jwt)
        try:
            user_id = me['datalake']['user_id']
        except Exception as e:
            self.logger.warning(f"For some reason we couldn't get the user id for the JWT : {me}")
            user_id = 'missing'

        from dataclasses import replace
        action_info = replace(action_info, user_id=user_id)

        payload = self._prepare_static_params_based_on(action_info)

        if action_info.user_id:
            payload.update({
                self.__params_mapping.get('user_id', 'missing'): action_info.user_id,
                self.__params_mapping.get('user_id_again', 'missing'): action_info.user_id,
            })

        if action_info.organisation_id:
            payload.update({
                self.__params_mapping.get('organisation_id', 'missing'):
                    action_info.organisation_id
            })

        if action_info.properties:
            payload.update(self._map_properties(action_info.properties))

        self.logger.debug(
            'AnalyticsEvent mapped from user email address to user ID',
            extra={'payload': payload}
        )

        return parse.urlencode(self._static_params + tuple(payload.items()))

    @staticmethod
    def _prepare_static_params_based_on(action_info):
        payload = {
            'an': action_info.application_name,
            'av': action_info.application_version,
            'cid': action_info.user_id,
            'ec': action_info.entity,
            'ea': action_info.action,
            'ev': action_info.result
        }

        return {k: v for k, v in payload.items() if v}

    def _map_properties(self, properties: dict):
        mapped_properties = {}
        for key in self.__params_mapping.keys():
            if properties.get(key, None) is not None:
                mapped_properties.update(
                    {self.__params_mapping[key]: properties[key]}
                )

        return mapped_properties



import base64
import logging
import os
from typing import Optional, NewType, Tuple, Dict, Any
from flask import request
from requests.models import CaseInsensitiveDict, Response

from ..authentication.auth import _pre_request_get_token, Unauthorized
from ..context.environments import _Environment
from ..providers.oidc_provider import JWTEncoded

logger = logging.getLogger(__name__)

########################################################################################################################
# The middleware relies on the application to inject a header in order to stay generic.
#
# Examples of how to bind the functions are shown in the 'Identity' (I) functions below, which do nothing.
# Albeit, any field of the original headers could be modified with the extracted JWT before sending to upstream.

# use to set injection of header modifier function
# defined by the type :: str, Dict -> Dict
HEADER_MODIFIER = NewType('HEADER_MODIFIER', str)
RESPONSE_MODIFIER = NewType('REQUEST_MODIFIER', str)


def identity_header_function(jwt: Optional[str], headers: Dict) -> Dict:
    # Add this to your flask app to bind this function, or provide your own
    #
    #     app.injector.binder.bind(
    #         HEADER_MODIFIER,
    #         to=lambda: identity_header_function,
    #         scope=request,
    #     )

    return headers


def identity_response_function(resp: Response) -> Tuple[Any, int, CaseInsensitiveDict]:
    # Add this to your flask app to bind this function, or provide your own
    #
    #     app.injector.binder.bind(
    #         RESPONSE_MODIFIER,
    #         to=lambda: identity_response_function,
    #         scope=request,
    #     )

    return resp.raw, resp.status_code, resp.headers


########################################################################################################################
# A flask middleware to be applied to the flask app - this resolves any possible latency or memory requirement
# caused by running as a separate server on the same pod
#
# This should be used as follows on creation of the flask app as follows:
#
#     app = Flask(__name__)
#     app.register_blueprint([views,...])
#
#     app.injector = Injector()
#     app.wsgi_app = ModifyHeaderMiddleware(app.wsgi_app)
#     app.wsgi_app.injector = app.injector
#

class ServiceMiddleware(object):
    def __init__(self, wsgi_app):
        self.wsgi_app = wsgi_app
        self.env = _Environment(os.environ["CATALOGUE_API_URL"])


    def __call__(self, environ: Dict, start_response):
        # before request to route, add some variables to the environ according to the app
        if not environ.get("HTTP_AUTHORIZATION"):
            start_response('401 Unauthorized', [('Content-Type', 'text/html')])
            return ["No Authorization header provided".encode()]

        jwt = ServiceMiddleware._resolve(self.env, "HTTP_AUTHORIZATION", environ)

        # lets remove this to make clear that we later set it
        environ.pop('HTTP_AUTHORIZATION', None)

        # this will interfere with the request
        environ.pop('HTTP_HOST', None)

        # this will likely set HTTP_AUTHORIZATION and any other headers it wants (prepended HTTP_)
        header_modifier_fn = self.injector.get(HEADER_MODIFIER)

        if header_modifier_fn:
            # application itself is requesting a header modification be run before sending to its routes
            environ = header_modifier_fn(jwt, environ)

        response_status = None
        response_headers = None

        def custom_start_response(status, environ=[]):
            nonlocal response_status
            nonlocal response_headers
            response_headers = environ
            response_status = status
            # after response from route, intercept the response (applicable only on plain out)
            # route_response = start_response(status, environ)
            # return route_response

        # go to the route with pre-adjusted environ, and response callback
        app_iter = self.wsgi_app(environ, custom_start_response)

        try:
            response_modifier = self.injector.get(RESPONSE_MODIFIER)
            if response_modifier:

                # recover the response as we still want to do some adjustment
                response_content_reread = b"".join(app_iter)
                dummy_response = Response()
                dummy_response._content = response_content_reread
                dummy_response.headers.update(response_headers)
                dummy_response.status_code = response_status

                # application itself is requesting a header modification be run before sending to its routes
                content, content_status, content_headers = response_modifier(dummy_response)
                start_response(str(content_status), list(dict(content_headers).items()))
                return [content]
            else:
                # if not modification needed, dont re-read, just return it* (this is an future)
                # based on the fact that you may still be writing/streaming the response in chunk
                # see https://stackoverflow.com/questions/16774952/wsgi-whats-the-purpose-of-start-response-function
                start_response(response_status, response_headers)
                return app_iter
        except Exception as e:
            start_response(str(e), [('Content-Type', 'text/html')])
            return [f"{str(e)}".encode()]

    @staticmethod
    def _resolve(env, auth_header="Authorization", use_environ=None) -> JWTEncoded:

        # Authorization schemes seen in the Authorization HTTP header
        bearer_prefix = "Bearer"
        ldap_basic_prefix = "Basic"
        aws_v4_signed_prefix = "AWS4-HMAC-SHA256"
        aws_s3_rest_prefix = "AWS"

        # Delimiter character which divides username and password in single field scenarios
        # Non-presence of the delimiter indicates that it can only be a JWT flow
        auth_credential_delimiter = "\\"

        user = None       # extracted username or JWT (if extracted)
        password = None   # extract password or "NOP" (if extracted)
        hint = None       # type of credential        (if can be inferred)

        if use_environ is None and request.authorization is not None:
            # for BI tools - username/password style forms
            # we know for a fact its a Credential: (App user/pass) or (PAT user/pass) or (JWT/nop), but not which
            if not use_environ:
                user = request.authorization.username
                password = request.authorization.password
            else:
                # shouldn't be receiving a middleware request with wsgi environ, this will raise Unauthorized later
                pass

        elif use_environ is not None or request.headers[auth_header] is not None:
            # we know this is either:
            # AMS4-HMAC-SHA256 <JWT>/... <User>:<Password>/.../.../  (v4 signing, used by AWS SDK [Spark/Boto etc])
            # AWS <JWT>/... <User>:<Password>/.../...                (s3 rest, used by AWS S3 over REST to spec. [CURL])
            # Bearer <JWT>                                           (Bearer, used by Datalake [S3-Proxy, Catalogue])
            # Basic <User>:<Password>                                (LDAP/Basic, [Presto])
            #
            # but we're not yet sure which, nor which Credential inhabits them.
            # check the prefix of the header

            # choose whether to look at the wsgi headers or use Flask request headers to get the Auth field
            if use_environ is None:
                auth = request.headers[auth_header]
            else:
                auth = use_environ[auth_header]

            # split the Authorization field into its parts
            space_idx = auth.find(" ")
            if space_idx != -1:
                auth_type_key = auth[0:space_idx].lower()
                auth_type_value = auth[space_idx+1:]
                extract_fn = None

                if auth_type_key == bearer_prefix.lower():
                    extract_fn = ServiceMiddleware._type_bearer
                elif auth_type_key == ldap_basic_prefix.lower():
                    extract_fn = ServiceMiddleware._type_ldap
                elif auth_type_key == aws_s3_rest_prefix.lower():
                    extract_fn = ServiceMiddleware._type_s3_rest
                elif auth_type_key == aws_v4_signed_prefix.lower():
                    extract_fn = ServiceMiddleware._type_v4_signature

                if extract_fn:
                    user, password, hint = extract_fn(auth_credential_delimiter, auth_type_value)
                else:
                    # doesn't look like a recognized auth scheme that we deal with
                    logger.warning(f"We cannot extract credentials for this type of authorization key {auth_type_key}")
            else:
                # doesn't look like an auth scheme
                pass

        if user is None:
            raise Unauthorized()

        return ServiceMiddleware._credentials_to_token(env, user, password, hint)

    @staticmethod
    def _credentials_to_token(env, user, password, hint):
        # we have this static method to make testing super easy to override
        return _pre_request_get_token(env, user, password, hint)

    @staticmethod
    def _split_to_jwt_or_credentials(
            auth_credential_delimiter, field_aws_access_key_id
    ) -> Tuple[str, str, Optional[str]]:
        if auth_credential_delimiter not in field_aws_access_key_id:
            # this MUST be a JWT, since there is no split character
            return field_aws_access_key_id, "NOP", "JWT"
        else:
            # we need to split into the respective values, split by the first we see
            # we still do not know which of the credential types it is however
            user, pasw = field_aws_access_key_id.split(auth_credential_delimiter, 1)
            return user, pasw, None

    @staticmethod
    def _type_ldap(auth_credential_delimiter, auth_type_value) -> Tuple[str, str, Optional[str]]:
        ldap_separator = ":"

        # only presto uses this, we know it is of the form (Basic <User>:<Password>) so we can split it
        # may be any one of the 3 credential types
        # we still do not know which of the credential types it is
        unencoded = base64.urlsafe_b64decode(auth_type_value).decode("ascii")
        user, pasw = unencoded.split(ldap_separator, 1)
        return user, pasw, None

    @staticmethod
    def _type_bearer(auth_credential_delimiter, auth_type_value) -> Tuple[str, str, Optional[str]]:
        # only datalake apps use bearer tokens, thus its a token of the form (Bearer <JWT>)
        # may only be JWT
        return auth_type_value, "NOP", "JWT"

    @staticmethod
    def _type_s3_rest(auth_credential_delimiter, auth_type_value) -> Tuple[str, str, Optional[str]]:
        aws_s3_rest_separator = ":"

        # only S3 REST, may be any one of the 3 credential types
        field_aws_access_key_id, _, _ = auth_type_value.strip().partition(aws_s3_rest_separator)
        return ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter, field_aws_access_key_id)

    @staticmethod
    def _type_v4_signature(auth_credential_delimiter, auth_type_value) -> Tuple[str, str, Optional[str]]:
        aws_v4_fields_separator = ","
        aws_v4_field_components_separator = "="
        aws_v4_credential_field = "Credential"
        aws_v4_credential_field_values_separator = "/"

        # only AWS V4 Signed use by s3proxy only, may be any one of the 3 credential types
        fields = auth_type_value.split(aws_v4_fields_separator)
        for field in fields:
            field_name, _, field_value = field.strip().partition(aws_v4_field_components_separator)

            if field_name == aws_v4_credential_field:
                field_aws_access_key_id, _, _ = field_value.partition(aws_v4_credential_field_values_separator)
                return ServiceMiddleware._split_to_jwt_or_credentials(auth_credential_delimiter, field_aws_access_key_id)

        # if Credential field is not in there, we need to return None, to hit Unauthorized
        return None, None, None


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import logging
import warnings

from deprecated.sphinx import deprecated

from platform_services_lib.lib.aspects.base_component import BaseComponent
from platform_services_lib.lib.context.urls import package_urls

logger = logging.getLogger(__name__)


class Package(BaseComponent):

    _KNOWN_FIELDS = {"name",
                     "description",
                     "keywords",
                     "topic",
                     "access",
                     "internalData",
                     "contractIds",
                     "termsAndConditions",
                     "derivedDataNotes",
                     "derivedDataRights",
                     "distributionNotes",
                     "distributionRights",
                     "internalUsageNotes",
                     "internalUsageRights",
                     "documentation",
                     "publisher",
                     "techDataOpsId",
                     "accessManagerId",
                     "managerId",
                     "intendedPurpose",
                     "isInternalWithinOrganisation"}
    """
    A mixin providing common package operations
    """

    @deprecated(version='1.17.0', reason="You should use `package=client.packages.get(<package_name>)` to get a package, followed by `package.datasets()`")
    def get_package_datasets(self, package_id, count=100):
        """
        Returns a list of all datasets registered under a package.

        :param str package_id: The id of the package.
        :param int count: Optional. Count of datasets to be returned. Defaults to 100.

        :returns: List of all datasets registered under the package.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                datasets = client.get_package_datasets(
                    package_id,
                    count=100
                )
        """
        warnings.warn(
            'This method is be deprecated and will be removed soon.'
            'Please use `client.datasets()` instead.',
            PendingDeprecationWarning
        )

        response = self.session.get(
            package_urls.v2_package_datasets.format(id=package_id),
            params={'page_size': count}
        )

        return self._DatasetFactory._from_v2_list_response(response.json())

    @classmethod
    def get_default_package_terms_and_conditions(cls, organisation_name: str):
        """
        Returns a string representing the default Terms And Conditions for packages created in DataLake for a given organisation.

        :returns: The default DataLake Terms And Conditions
        :rtype: str
        """
        # Scott: please do not deprecate this function. It is used by Excel users.
        if organisation_name == 'IHS Markit':
            return ('By submitting this Data request and checking the "Accept Terms and Conditions" '
                'box, you acknowledge and agree to the following:\n'
                '\n'
                '* To promptly notify the relevant Access Manager/Producer of your intended use '
                'of the Data;\n'
                '* To obtain the terms and conditions relevant to such use for such Data from '
                'the Producer;\n'
                '* To distribute such terms and conditions to each member of your '
                'Consumer Group who may use the Data;\n'
                '* To use the Data solely for such intended use, subject to such terms and '
                'conditions;\n'
                '* To ensure that the Data is only accessed by members of your Consumer Group, '
                'and only used by such members for such intended use, subject to such terms and '
                'conditions;\n'
                '* To adhere to any additional requests of Producer with respect to the Data '
                '(including but not limited to ceasing use of the Data and deleting the Data, '
                'and ensuring other members of the Consumer Group do so, upon revocation of your '
                'license by Producer).\n'
                '\n'
                'Please refer to the <a href="/terms-of-use" target="_blank">EULA</a> for any '
                'defined terms used above. '
                'The <a href="/terms-of-use" target="_blank">EULA</a> '
                'is the document you agreed to adhere to by accessing the Lake.')
        else:
            return ''



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import inject
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function):  # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper
except:
    pass

from ..analytics_service.model import AnalyticsEvent
from ..ga_service.ga_service import GoogleAnalyticsService


class AnalyticsService:

    @inject
    def __init__(self, analytics_service: GoogleAnalyticsService):
        self._analytics_service = analytics_service

    def process(self, action_info: AnalyticsEvent) -> None:
        if not action_info:
            return

        body = self._analytics_service.create_message_body(action_info)
        self._analytics_service.send_message(body)



import inspect
from functools import wraps
from typing import List, Optional

from . import _make_function_call
from .analytics_aspect import AnalyticsAspect
from .logging_aspect import LoggingAspect
from ..services.dlc_attributes_dict import AttributesDict


def log_public_functions_calls_using(decorators, class_fields_to_log=None, excluded_methods: Optional[List[str]]=None):
    """
    This decorates all functions in a Model/Object with a logging function, the equivalent to ComponentsAspectWrapper()
    This method should outside the class, to decorate it, applying named aspects to all its methods*
    e.g.

    log_public_functions_calls_using(
        [analytics_decorator, logging_decorator],
        class_fields_to_log=['dataset_id']
    )(StructuredDatasetModel)

    """
    if not excluded_methods:
        excluded_methods = []
    if not class_fields_to_log:
        class_fields_to_log = []

    def decorate(cls):
        # SDK specific inspection to exclude functions of underlying base type (AttributesDict).
        # Use `excluded_methods` for normal classes where a specific method other than _ or __ ones needs excluding
        functions_to_exclude = inspect.getmembers(AttributesDict, inspect.isfunction)
        functions_to_decorate = [
            func for func in inspect.getmembers(cls, inspect.isfunction)
            if func not in functions_to_exclude
               and not func[0].startswith('_')
               and not func[0].startswith('__')
               and func[0] not in excluded_methods
        ]
        for function_meta in functions_to_decorate:
            function_name = function_meta[0]
            # add the decorator to cls.function_name
            for decorator in decorators:
                setattr(
                    cls,
                    function_name,
                    decorator(getattr(cls, function_name),
                              class_fields_to_include=class_fields_to_log)
                )
        return cls
    return decorate


def analytics_decorator(function, class_fields_to_include):
    aspect = [AnalyticsAspect()]

    @wraps(function)
    def function_wrapper(target, *args, **kwargs):
        return _make_function_call(aspect, target, function, args, kwargs, class_fields_to_include)

    return function_wrapper


def logging_decorator(function, class_fields_to_include):
    aspect = [LoggingAspect()]

    @wraps(function)
    def function_wrapper(target, *args, **kwargs):
        return _make_function_call(aspect, target, function, args, kwargs, class_fields_to_include)

    return function_wrapper



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import datetime
import glob
import logging
import multiprocessing
import os
import posixpath
import warnings
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
from typing import Dict, List
from urllib.parse import urlparse, ParseResult
import botocore
from boto3.session import Session
from dateutil import tz
from wrapt import ObjectProxy

from dli.components import _AttributeAdapter, object_to_namedtuple, to_snake_case_keys
from dli.models.dataset_model import DatasetModel

from platform_services_lib.lib.aspects.base_component import BaseComponent
from platform_services_lib.lib.services.exceptions import (
    DownloadFailed, DatalakeException, S3FileDoesNotExist,
    InsufficientPrivilegesException, S3BucketDoesNotExist
)
from platform_services_lib.lib.context.urls import datafile_urls

THREADS = multiprocessing.cpu_count()
logger = logging.getLogger(__name__)


class File(ObjectProxy):
    """
    This is only to mantain backwards compatibility with S3FS.
    Ideally we want to deprecate getting files as anything but
    context managers. This provides both the context manager
    and function that returns a files like object
    """
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__wrapped__.close()


class S3DatafileWrapper:
    def __init__(self, datafile, client):
        # we want to override the files property
        # it isn't really needed for py3 but it is in py2
        self.__dict__ = {
            k: v for k,v in datafile.items() if k != "files"
        }
        self._datafile = datafile
        self._client = client

    def as_dict(self):
        return dict(self._datafile)

    def _asdict(self):
        return self.as_dict()

    def __getitem__(self, path):
        """
        :param str path: The path of the file to retrieve from s3.
        :returns: A S3File instance
        """
        return self.open_file(path)

    def open_file(self, path):
        """
        Helper method to load an specific file inside a datafile
        without having to download all of it.

        :param str path: The path of the file to retrieve from s3.
        :returns: A S3File instance that can be used as a normal python file
        """
        bucket, _, key = path.partition('/')

        try:
            bucket = self._client.s3_resource.Bucket(bucket)
            response = bucket.Object(key).get()['Body']
            return File(response)
        except (
            self._client.s3_resource.meta.client.exceptions.NoSuchKey,
            self._client.s3_resource.meta.client.exceptions.NoSuchBucket
        ) as e:
            raise S3FileDoesNotExist(path)

    def __repr__(self):
        return str(dict(self._datafile))

    @property
    def files(self):
        """
        Lists all S3 files in this datafile, recursing on folders (if any)
        """
        files = self._datafile["files"]
        if not files:
            return []

        bucket = urlparse(files[0]['path']).netloc

        prefix = os.path.commonprefix([
            urlparse(file_['path']).path for file_ in files
        ]).lstrip('/')

        return [
            '{}/{}'.format(bucket, obj['Key']) for obj in
            self._client._list_objects(bucket, prefix)
        ]


class S3Client:
    """
    A wrapper client providing util methods for boto
    """

    def __init__(self, boto_session):
        # not called directly other than for testing purposes, called from from_dataset[id]
        self.boto_session = boto_session

        self.s3_resource = self.boto_session.resource(
            's3', endpoint_url=os.environ.get('S3_URL'),
        )

        self.s3_client = self.s3_resource.meta.client

    @classmethod
    def from_dataset(cls, client, dataset_id) -> 'S3Client':
        def _credentials():
            keys = client.get_s3_access_keys_for_dataset(dataset_id)
            expiry_time = (
                # expiring the token after an hour. It's conservative
                # but it should just be refreshed. TODO add expiry time
                # to the get_s3_access_keys_for_dataset endpoint
                datetime.datetime.now().replace(tzinfo=tz.tzlocal()) + timedelta(hours=1)
            ).isoformat()

            return dict(
                access_key=keys.access_key_id,
                secret_key=keys.secret_access_key,
                token=keys.session_token,
                expiry_time=expiry_time
            )

        session_credentials: botocore.credentials.RefreshableCredentials = \
            botocore.credentials.RefreshableCredentials.create_from_metadata(
                metadata=_credentials(),
                refresh_using=_credentials,
                method='NOOP'
            )

        session: Dict = botocore.session.get_session()
        # Extremely dirty but it seems to be the 'blessed' way of doing this.
        session._credentials: botocore.credentials.RefreshableCredentials = session_credentials
        return cls(Session(botocore_session=session))

    def _list_objects(self, bucket, prefix):
        # as it turns out the high level boto client is rather expensive
        # when creating objects en-masse. This is at least 50% faster.
        # This is also the exact way that the aws cli lists files.
        paginator = self.s3_client.get_paginator(
            'list_objects_v2'
        )

        paging_args = {'Bucket': bucket, 'Prefix': prefix}
        iterator = paginator.paginate(**paging_args)

        for response_data in iterator:
            for obj in response_data.get('Contents'):
                yield obj

    def download_files_from_s3(self, s3_location, destination, flatten=False):
        """
        Helper function to download a file as a stream from S3
        :param str s3_location: required, example: s3://bucket_name/prefix_part1/prefix_part2/file_name.txt

        :param str destination: required. The path on the system, where the
            files should be saved. must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :returns: a list of the paths where the files were downloaded
            to e.g. ['path/git /package/dataset/as_of_date=2019-10-16
            /file.csv.gz'].
            with flatten=True e.g. ['path/file.csv.gz']

        :rtype: str
        """
        if os.path.isfile(destination):
            raise NotADirectoryError(destination)

        parse_result = urlparse(s3_location)
        bucket = parse_result.netloc
        # Removes forward slash
        download_path = parse_result.path.lstrip('/').rstrip('/')

        def _download(obj_summary):
            # Ignore 'folders'
            key = obj_summary['Key']
            destination_key = key
            if not key.endswith('/'):

                def flatten_s3_file_path(s3_path):
                    from urllib.parse import urlparse
                    import ntpath
                    path = urlparse(s3_path)[2]  # [scheme, netloc, path, ...]
                    head, tail = ntpath.split(path)
                    file = tail or ntpath.basename(head)
                    return file

                if flatten:
                    destination_key = flatten_s3_file_path(s3_path=key)

                destinatation_path = os.path.join(destination, destination_key)
                subfolder = os.path.dirname(destinatation_path)
                os.makedirs(subfolder, exist_ok=True)
                logger.info('Downloading File %s to %s', key, destinatation_path)
                self.s3_client.download_file(bucket, key, destinatation_path)
                return destinatation_path

        keys = list(self._list_objects(bucket, download_path))
        threads = min(THREADS, len(keys))
        with ThreadPoolExecutor(max_workers=threads) as executor:
            destination_path = list(executor.map(_download, keys))
        return destination_path

    def upload_files_to_s3(self, files, s3_location) -> List[Dict]:
        """
        Upload multiple files to a specified s3 location. The basename of the
        `files` will be preserved.

        :param List files: An list of filepaths
        :param str s3_location: Path to destination directory in S3, file will be
                 stored under <s3_location><filename>
        :returns: List of path to the files in S3
        :rtype: List[str]
        """

        bucket_name, *prefix = s3_location.partition('/')
        try:
            # Verify that the bucket exists to avoid support tickets of people trying to upload
            # to buckets with wrong spellings or buckets that do not exist.
            # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/migrations3.html#accessing-a-bucket
            bucket = self.s3_client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = e.response['Error']['Code']
            if error_code == '404':
                raise S3BucketDoesNotExist(bucket_name) from e
            elif error_code == '403':
                # AWS permissions are setup in strange ways, like in the case that the user
                # has permission to upload to s3 but cannot call read operations such as
                # `head_bucket`. QA had this issue, so it breaks the code that checks if the
                # bucket exists before uploading.
                bucket = self.s3_resource.Bucket(bucket_name)
            else:
                # All other error_codes should bubble up the original exception.
                raise

        prefix = ''.join(prefix)

        def _files_to_upload_flattened(files):
            files_to_upload = []

            for file in files:
                if not os.path.exists(file):
                    raise Exception('File / directory specified (%s) for upload does not exist.' % file)

                if os.path.isfile(file):
                    logger.info("detected file: %s", file)
                    files_to_upload.append({'file': file,
                                            's3_suffix': os.path.basename(file)})
                elif os.path.isdir(file):
                    logger.info("detected directory: %s", file)
                    all_contents = glob.glob(os.path.join(file, "**/*"), recursive=True)
                    fs = [f for f in all_contents if os.path.isfile(f)]

                    for f in fs:
                        files_to_upload.append({
                            'file': f,
                            's3_suffix': os.path.relpath(f, file).replace("\\", "/")
                        })

            return files_to_upload

        def upload(file, s3_location) -> Dict:
            file_path = file['file']
            s3_suffix = file['s3_suffix']
            target = os.path.join(prefix, s3_suffix)

            logger.info("Uploading %s to %s", file_path, target)

            try:
                self.s3_client.upload_file(
                    file_path, bucket_name, target.lstrip('/')
                )
            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == 'AccessDenied':
                    raise InsufficientPrivilegesException() from e

                raise e

            logger.info(".. %s uploaded successfully.", file_path)

            return {
                "path": ParseResult(
                    scheme='s3',
                    netloc=bucket_name,
                    path=target,
                    params='', query='', fragment=''
                ).geturl(),
                "size": os.path.getsize(file_path)
            }

        files_to_upload = _files_to_upload_flattened(files)
        result = []
        for file in files_to_upload:
            uploaded = upload(file, s3_location)

            result.append(uploaded)

        return result


# utilities above


class Datafile(BaseComponent):

    def register_datafile_metadata(
        self,
        dataset_id,
        name,
        files,
        data_as_of,
        original_name=None
    ):
        """
        Submit a request to create a new datafile under a specified dataset in the Data Catalogue. This function
        ``WILL NOT`` upload files

        Datafile is an instance of the data within a dataset, representing a snapshot of the data at the time of
        publishing. A dataset can be composed by one or more related files that share a single schema. of related
        datafiles. It provides user with metadata required to consume and use the data.

        See description for each parameter, and whether they are optional or mandatory.

        :param str dataset_id: Dataset ID to which the datafile belongs to.
        :param str name: A descriptive name of a datafile. It should be unique within the dataset.
        :param list[dict] files: List of file dictionaries. A file dictionary will contain the full file path
                                and size (optional) as items.
        :param str data_as_of: The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.
        :param str original_name: Optional. Name of the data uploaded into the data lake.

        :returns: Registered datafile
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.register_datafile_metadata(
                    dataset_id,
                    name="My Datafile",
                    files=[{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}],
                    data_as_of="2018-05-28"
                )
        """
        ds: Dict[str, DatasetModel] = self.datasets(search_term=[f"id={dataset_id}"])
        num_matches = len(ds)
        if num_matches == 0:
            raise DatalakeException(f"No such dataset ID `{dataset_id}`")
        elif num_matches > 1:
            # This should not be possible, but let's cover this scenario incase Caatlogue returns
            # bad data.
            raise DatalakeException(f"More than one dataset with ID `{dataset_id}`")
        else:
            dataset: DatasetModel = list(ds.values())[0]

        if not files:
            raise DatalakeException("No files to register have been provided.")

        fields = {
            'datasetId': dataset.dataset_id,
            'name': name,
            'originalName': original_name,
            'dataAsOf': self._ensure_iso_date(data_as_of),
            'files': files,
        }

        payload = {k: v for k, v in fields.items() if v is not None}

        response_dict = self.session.post(
                    datafile_urls.datafiles_index, json=payload
                ).json()
        obj = type(response_dict['class'][0], (_AttributeAdapter,), to_snake_case_keys(response_dict["properties"]))()
        return obj

    def register_s3_datafile(
        self,
        dataset_id,
        name,
        files,
        s3_prefix,
        data_as_of,
        original_name=None
    ):
        """
        Submit a request to create a new datafile under a specified dataset in the Data Catalogue.
        This function will perform an upload of the files to S3 data store.

        Datafile is an instance of the data within a dataset, representing a snapshot of the data at the time of publishing.
        A dataset can be composed by one or more related files that share a single schema. of related datafiles.
        It provides user with metadata required to consume and use the data.

        See description for each parameter, and whether they are optional or mandatory.

        :param str dataset_id: Dataset ID to which the datafile belongs to.
        :param str name: A descriptive name of a datafile. It should be unique within the dataset.
        :param list[str] files: Path of the files or folders to register.
        :param str s3_prefix: location of the files in the destination bucket. Cannot end with slash ("/")
        :param str data_as_of: The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.
        :param str original_name: Optional. Name of the data uploaded into the data lake.

        :returns: Registered datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.register_s3_datafile(
                    dataset_id=dataset_id,
                    name="My datafile",
                    files=["./test_sandbox/samples/data/AAPL.csv", "./test_sandbox/samples/data/MSFT.csv"],
                    s3_prefix="quotes/20180518",
                    data_as_of="2018-05-28"
                )
        """
        ds: Dict[str, DatasetModel] = self.datasets(search_term=[f"id={dataset_id}"])
        num_matches = len(ds)
        if num_matches == 0:
            raise Exception(f"No such dataset ID `{dataset_id}`")
        elif num_matches > 1:
            # This should not be possible, but let's cover this scenario incase Caatlogue returns
            # bad data.
            raise Exception(f"More than one dataset with ID `{dataset_id}`")
        else:
            dataset: DatasetModel = list(ds.values())[0]

        if not files:
            raise Exception("No files to register have been provided.")

        # upload files
        location = None if hasattr(dataset, 'location') is False else dataset.location
        if not location:
            raise Exception(
                f"No access to specified dataset.")

        if location.type != 'S3':
            raise Exception(
                "Only datasets backed on S3 are supported. use register_datafile_metadata instead.")

        bucket = location.bucket
        if not bucket:
            raise Exception(
                "Dataset location is S3, however, "
                "there is no bucket associated with the dataset {}".format(
                    dataset_id)
            )

        s3_location = "{}/{}".format(bucket, s3_prefix)
        uploaded = self._process_s3_upload(files, s3_location, dataset_id)


        return self.register_datafile_metadata(
            dataset_id,
            name=name,
            files=uploaded,
            original_name=original_name,
            data_as_of=self._ensure_iso_date(data_as_of),
        )

    def edit_datafile_metadata(
        self,
        datafile_id,
        name=None,
        original_name=None,
        data_as_of=None,
        files=None
    ):
        """
        Edits metadata for an existing datafile.
        This function WILL NOT upload files.
        Fields passed as ``None`` will retain their original value.

        :param str datafile_id: The id of the datafile we want to modify.
        :param str name: Optional. Name of the datafile.
        :param str original_name: Optional. Original Name for the datafile.
        :param str data_as_of: Optional. The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.
        :param list[dict] files: Optional. List of file dicts. A file dict will contain file path and size(optional) as items.

        :returns: Registered datafile
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                updated_datafile = client.edit_datafile_metadata(
                    datafile_id,
                    name="My Datafile",
                    files=[{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
                )
        """

        datafile = self._get_datafile(datafile_id)

        fields = {
            'datasetId': datafile.datasetId,
            'name': name,
            'originalName': original_name,
            'dataAsOf': self._ensure_iso_date(data_as_of),
            'files': files
        }

        # clean up any unknown fields, and update the entity
        datafile_as_dict = datafile.as_dict()

        for key in list(datafile_as_dict.keys()):
            if key not in fields:
                del datafile_as_dict[key]

        payload = {k: v for k, v in fields.items() if v is not None}
        datafile_as_dict.update(payload)

        # perform the update and return the resulting entity
        response_dict = self.session.put(
            datafile_urls.datafiles_instance.format(id=datafile_id),
            json=datafile_as_dict
        ).json()
        obj = type(response_dict['class'][0], (_AttributeAdapter,), to_snake_case_keys(response_dict["properties"]))()
        named_tuple_obj = object_to_namedtuple(obj)
        return named_tuple_obj

    def delete_datafile(self, datafile_id):
        """
        Marks a datafile as deleted.

        :param str datafile_id: the unique id for the datafile we want to delete.

        :returns: None

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                client.delete_datafile(datafile_id)
        """
        response = self.session.delete(
            datafile_urls.datafiles_instance.format(id=datafile_id)
        )

    def get_s3_datafile(self, datafile_id):
        """
         .. deprecated:: 1.8.0

            This method is be deprecated and will be removed in release 1.18.0.
            Please change your workflow to instead get the dataset and then call download on
            that dataset. Filters can be provided to target data within the S3 partitions such
            'as `as_of_date` e.g.

            .. code-block:: python

                    dataset = client.datasets.get(
                        dataset_short_code='ExampleDatasetShortCode',
                        organisation_short_code='IHSMarkit'
                    )"

                    dataset.download(partitions=['as_of_date>=2020-01-01','country=US'])

            This dataset `download` call will download using the contents that is on S3.

        Fetches an S3 datafile providing easy access to directly
        stream/load files without the need of downloading them.

        If the datafile is not stored in S3, or you don't have access to it
        then an error will be displayed.

        :param str datafile_id: The id of the datafile we want to load

        :returns: a datafile that can read files from S3
        :rtype: dli.client.components.datafile.S3DatafileWrapper

        .. code-block:: python

                client = dli.connect()
                datafile = client.get_s3_datafile(datafile_id)
                with datafile.open_file("path/to/file/in/datafile") as f:
                    f.read() # do something with the file

                # or if you want a pandas dataframe created from it you can
                pd.read_csv(datafile.open_file("path/to/file/in/datafile"))

                # you can see all the files in the datafile by calling `files`
                datafile.files  # displays a list of files in this datafile

        """
        warnings.warn(
            'This method is be deprecated and will be removed in release 1.18.0. '
            '\nPlease change your workflow to instead get the dataset and then call download on '
            'that dataset. Filters can be provided to target data within the S3 partitions such '
            'as `as_of_date` e.g. '
            "\n\n    dataset = client.datasets.get(dataset_short_code='ExampleDatasetShortCode', "
            "organisation_short_code='IHSMarkit')"
            "\n\n    dataset.download(partitions=['as_of_date>=2020-01-01','country=US'])"
            '\n\nThis dataset `download` call will download using the contents that is on S3.',
            PendingDeprecationWarning
        )

        datafile = self.get_datafile(datafile_id)
        s3_access = S3Client.from_dataset(self, datafile.dataset_id)
        return S3DatafileWrapper(datafile._asdict(), s3_access)

    def download_datafile(self, datafile_id, destination, flatten=False):
        """
         .. deprecated:: 1.17.0
            This method is be deprecated and will be removed in release 1.18.0.
            Please change your workflow to instead get the dataset and then call download on
            that dataset. Filters can be provided to target data within the S3 partitions such
            'as `as_of_date` e.g.

            .. code-block:: python

                    dataset = client.datasets.get(
                        dataset_short_code='ExampleDatasetShortCode',
                        organisation_short_code='IHSMarkit'
                    )"

                    dataset.download(partitions=['as_of_date>=2020-01-01','country=US'])

            This dataset `download` call will download using the contents that is on S3.

        Helper function that downloads all files
        registered in a datafile into a given destination.

        This function is only supported for data-lake managed s3 buckets,
        otherwise an error will be displayed.

        Currently supports:
          - s3

        :param str datafile_id: The id of the datafile we want to download files from.

        :param str destination: required. The path on the system, where the
            files should be saved. must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: Optional. The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :returns: a list of the paths where the files were downloaded
            to e.g. ['path/git /package/dataset/as_of_date=2019-10-16
            /file.csv.gz'].
            with flatten=True e.g. ['path/file.csv.gz']

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                client.download_datafile(datafile_id, destination)
        """
        warnings.warn(
            'This method is be deprecated and will be removed in release 1.18.0. '
            '\nPlease change your workflow to instead get the dataset and then call download on '
            'that dataset. Filters can be provided to target data within the S3 partitions such '
            'as `as_of_date` e.g. '
            "\n\n    dataset = client.datasets.get(dataset_short_code='ExampleDatasetShortCode', "
            "organisation_short_code='IHSMarkit')"
            "\n\n    dataset.download(partitions=['as_of_date>=2020-01-01','country=US'])"
            '\n\nThis dataset `download` call will download using the contents that is on S3.',
            PendingDeprecationWarning
        )
        # get the s3 keys
        # this requires access to be granted
        datafile = self._get_datafile(datafile_id)
        s3_access = S3Client.from_dataset(self, datafile.datasetId)
        # for each file/folder in the datafile, attempt to download the file
        # rather than failing at the same error, keep to download as much as
        # possible and fail at the end.
        failed = []
        files = [f["path"] for f in datafile.files]
        filepaths = []
        for file in files:
            try:
                download_path = s3_access.download_files_from_s3(
                    s3_location=file,
                    destination=destination,
                    flatten=flatten,
                )
                filepaths.extend(download_path)
            except Exception:
                logger.exception(
                    "Failed to download file `%s` from datafile `%s`", file, datafile_id)
                failed.append(file)

        if failed:
            raise DownloadFailed(
                "Some files in this datafile could not be downloaded, "
                "see logs for detailed information. Failed:\n%s"
                % "\n".join(failed)
            )
        return filepaths

    def _process_s3_upload(self, files, s3_location, dataset_id):
        s3_client = S3Client.from_dataset(self, dataset_id)
        # Ensure trailing slash is included if missing
        s3_location = posixpath.join(s3_location, '')
        return s3_client.upload_files_to_s3(files, s3_location)

    def get_datafile(self, datafile_id):
        """
        Fetches datafile metadata for an existing datafile.

        :param str datafile_id: the unique id of the datafile we want to fetch.

        :returns: The datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                datafile = client.get_datafile(datafile_id)
        """
        datafile = self._get_datafile(datafile_id, snake_case=True)

        return object_to_namedtuple(datafile)

    def _get_datafile(self, datafile_id, snake_case: bool = False):
        response_dict = self.session.get(
            datafile_urls.datafiles_instance.format(id=datafile_id)
        ).json()
        if snake_case:
            obj = type(response_dict['class'][0], (_AttributeAdapter,), to_snake_case_keys(response_dict["properties"]))()
        else:
            obj = type(response_dict['class'][0], (_AttributeAdapter,), response_dict["properties"])()
        return obj

    def add_files_to_datafile(self, datafile_id, s3_prefix, files):
        """
        Upload files to existing datafile.

        :param str datafile_id: The id of the datafile to be updated.
        :param str s3_prefix: Location for the files in the destination s3 bucket. Cannot end with slash ("/").
        :param list[str] files: List of files to be added to the datafile.

        :returns: The updated datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                datafile_updated_with_additional_files = client.add_files_to_datafile(
                                                              datafile_id=datafile_id,
                                                              s3_prefix="quotes/20180518",
                                                              files=["./data/AAPL.csv", "./data/MSFT.csv"],
                                                        )
        """
        datafile = self.get_datafile(datafile_id)
        dataset_id = datafile.dataset_id

        ds: Dict[str, DatasetModel] = self.datasets(search_term=[f"id={dataset_id}"])
        num_matches = len(ds)
        if num_matches == 0:
            raise Exception(f"No such dataset ID `{dataset_id}`")
        elif num_matches > 1:
            # This should not be possible, but let's cover this scenario incase Caatlogue returns
            # bad data.
            raise Exception(f"More than one dataset with ID `{dataset_id}`")
        else:
            dataset: DatasetModel = list(ds.values())[0]

        location = None if hasattr(dataset, 'location') is False else dataset.location
        if not location:
            raise Exception("No access to dataset.")

        if dataset.location.type != "S3":
            raise Exception("Can not upload files to non-S3 datasets.")

        s3_location = "{}/{}".format(dataset.location.bucket, s3_prefix)
        uploaded_files = self._process_s3_upload(
            files,
            s3_location,
            datafile.dataset_id
        )

        if datafile.files:
            uploaded_files.extend(datafile.files)

        return self.edit_datafile_metadata(datafile_id, files=uploaded_files)

    @staticmethod
    def _ensure_iso_date(date):
        if date:
            return datetime.datetime.strptime(date, "%Y-%m-%d").strftime("%Y-%m-%d")
        return date




import warnings
from typing import List

from dli.models.instance_model import InstanceModel

from platform_services_lib.lib.adapters.pagination_adapters import Paginator
from platform_services_lib.lib.context.urls import dataset_urls
from platform_services_lib.lib.aspects.decorators import analytics_decorator, logging_decorator, log_public_functions_calls_using
from platform_services_lib.lib.services.dlc_attributes_dict import AttributesDict


class InstanceModelCollection(AttributesDict):

    def __init__(self, dataset=None, page_size=25):
        self._dataset = dataset
        self._paginator = Paginator(
            dataset_urls.datafiles.format(id=self._dataset.id),
            self._client._Instance, self._client._Instance._from_v1_entity,
            page_size=page_size,
        )

    def latest(self) -> InstanceModel:
        """:return: The latest instance."""
        response = self._client.session.get(
            dataset_urls.latest_datafile.format(id=self._dataset.id)
        ).json()

        return self._client._Instance._from_v1_entity(response)

    def all(self) -> List[InstanceModel]:
        """:return: All the instances."""
        warnings.warn(
            'The result of calling `.all` will be cached. If you want fresh '
            'results the next time you call `.all`, then please re-create the '
            'dataset variable before calling `.all`.'
        )
        yield from self._paginator

log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['_dataset.dataset_id']
)(InstanceModelCollection)




import json
import logging
from requests import HTTPError

from ..services.exceptions import InterfaceResourceUnavailable

logger = logging.getLogger(__name__)


class ExternalService:
    @classmethod
    def _service_response_error(cls, err_msg: str):
        """
        Request function for handling errors requesting
        interface resources.
        """
        def _hook(response, *args,  **kwargs):
            try:
                response.raise_for_status()
            except HTTPError as e:
                error_data = None

                try:
                    error_data = response.json()
                except json.decoder.JSONDecodeError:
                    error_data = response.text

                logger.warning(
                    f'Error accessing external service {response.request}',
                    extra={
                        'response': response.status_code,
                        'error_data': error_data,
                        'url': response.request.url
                    }
                )

                raise InterfaceResourceUnavailable(
                    details=err_msg,
                    status=e.response.status_code
                ) from e

        return _hook

    @classmethod
    def _make_hook(cls, err_msg: str):
        return {
            'response': [cls._service_response_error(err_msg)]
        }


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import ntpath
import os
from collections import UserDict
from typing import Mapping
from urllib.parse import urlparse


class AttributesDict(UserDict):

    def __init__(self, *args, **kwargs):
        # recursively provide the rather silly attribute access
        data = {}

        for arg in args:
            data.update(arg)

        data.update(**kwargs)

        for key, value in data.items():
            if isinstance(value, Mapping):
                self.__dict__[key] = AttributesDict(value)
            else:
                self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        attributes = ' '.join([
            '{}={}'.format(k, v) for k,v in self.__dict__.items()
            if not k.startswith('_')
        ])

        return "{}({})".format(self.__class__.__name__, attributes)


def print_model_metadata(model: AttributesDict):
    for (k, v) in vars(model).items():
        if k == "documentation" and type(v) is str:
            val = v[0:100].replace('\n', ' ').replace('\r', '') + "..." \
                if len(v) > 100 else v
        else:
            val = v

        if not (hasattr(v, '__dict__') or k.startswith("_")):
            print(f"{k}: " + f"{val}")


def flatten_s3_file_path(s3_path):
    path = urlparse(s3_path)[2]  # [scheme, netloc, path, ...]
    head, tail = ntpath.split(path)
    file = tail or ntpath.basename(head)
    return file


def get_or_create_os_path(s3_path: str, to: str, flatten: bool) -> str:
    if flatten:
        destination_key = flatten_s3_file_path(s3_path)
    else:
        destination_key = urlparse(s3_path).path.lstrip('/')

    to_path = os.path.join(
        to, os.path.normpath(destination_key)
    )

    to_path = os.path.abspath(to_path)

    if len(to_path) > 259 and os.name == 'nt':
        raise Exception(f"Apologies {s3_path} can't be downloaded "
                        f"as the file name would be too long. You "
                        f"may want to try calling again with "
                        f"Instance.download(flatten=True), which "
                        f"will put the file in a directory of your choice")

    else:
        directory, _ = os.path.split(to_path)
        os.makedirs(directory, exist_ok=True)

    return to_path


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import inspect


class LoggingAspect:

    # N.B. The contract is that the wrapped_object HAS_A logger property

    @staticmethod
    def invoke_pre_call_aspects(wrapped_object, metadata):
        if getattr(wrapped_object, 'logger', None):
            try:
                wrapped_object.logger.debug('Client Function Called', extra=metadata)
            except Exception as e:
                wrapped_object.logger.exception(
                    'Error while invoking pre-call aspects.', e
                )

    @staticmethod
    def invoke_post_call_aspects(wrapped, metadata):
        pass

    @staticmethod
    def invoke_after_exception_aspects(wrapped_object, metadata, exception):
        # If the object contains a `strict` boolean and it is set to
        # True, then print out the exception message and a full stack
        # trace.
        # DEFAULTED to True to match the current behaviour. This default
        # can be changed in a later release. When we change the default
        # we will have to update the tests to explicitly set strict to
        # True to maintain the previous test behaviour.
        if getattr(wrapped_object, 'logger', None):
            try:
                wrapped_object.logger.exception(
                    'Unhandled Exception', stack_info=False, extra={
                        'locals': inspect.trace()[-1][0].f_locals
                    })
            except Exception as e:
                wrapped_object.logger.exception(
                    'Error while invoking pre-call aspects.', e
                )


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import copy
import json
import os
from requests_toolbelt import MultipartEncoder

from dli.components import _AttributeAdapter, to_snake_case_keys

from platform_services_lib.lib.aspects.base_component import BaseComponent
from platform_services_lib.lib.services.exceptions import CatalogueEntityNotFoundException
from platform_services_lib.lib.context.urls import dataset_urls

TEN_MEGABYTES = 10485760


class DatasetBuilder:
    """
    Dataset is grouping of related datafiles. It provides user with metadata required to consume and use the data.

    This builder object sets sensible defaults and exposes
    helper methods on how to configure its storage options..

    :param str package_id: Package ID to which the dataset belongs to.
    :param str name: A descriptive name of the dataset. It should be unique within the package the dataset
                    belongs to.
    :param str description: A short description of the dataset.
    :param str content_type: A way for the data steward to classify the type of data in the dataset
                        (e.g. Structured or Unstructured).
    :param str data_format: The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `TSV`,`XML`, `Other`.
    :param str publishing_frequency: The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
    :param list[str] taxonomy: A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
    :param str short_code: Optional. A short code for the dataset, should be an alphanumerical camel-cased string.
                        If not provided, a short_code will be generated based on the name.
    :param list[str] keywords: Optional. User-defined list of separated keywords that can be used to find this dataset
                            through the search interface.
    :param str naming_convention: Optional. Key for how to read the dataset name.
    :param str documentation: Optional. Documentation about this dataset in markdown format.
    :param str load_type: Optional. Whether each datafile in this dataset should be considered as a full version of a
                        dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`
    :param str has_datafile_monitoring: Optional. Notify when latest instance is missing
    :param bool details_visible_to_lite_users: Optional. Default value "False".
    :param str data_preview_type: Optional enum of NONE STATIC LIVE. 'NONE' means there is no datapreview. STATIC means to use the file provided. LIVE means to use the latest datafile.
    :param file sample_data: Optional file. File like object. Used in conjuntion with data_preview_type.
    """

    @staticmethod
    def _create_dliv1_payload(
        package_id=None,
        name=None,
        description=None,
        content_type=None,
        data_format=None,
        publishing_frequency=None,
        taxonomy=None,
        keywords=None,
        naming_convention=None,
        documentation=None,
        load_type=None,
        has_datafile_monitoring=None,
        details_visible_to_lite_users=None,
        **kwargs
    ):
        return {
            "packageId": package_id,
            "name": name,
            "description": description,
            "keywords": keywords,
            "contentType": content_type,
            "location": None,
            "dataFormat": data_format,
            "publishingFrequency": publishing_frequency,
            "namingConvention": naming_convention,
            "documentation": documentation,
            "taxonomy": taxonomy,
            "loadType": load_type,
            "has_datafile_monitoring": has_datafile_monitoring,
            "details_visible_to_lite_users": details_visible_to_lite_users,
        }

    def __init__(self, location_builder=None, **kwargs):
        self._data = dict(kwargs)

        self._data['sample_data'] = self._data.get('sample_data')
        if self._data.get('sample_data'):
            self._data['data_preview_type'] = (
                    self._data.get('data_preview_type') or 'STATIC'
            )
            self._validate_preview_type_static(
                self._data.get('data_preview_type')
            )

        # this is just for DLI v1 TODO add deprecated to it.
        self.payload = self._create_dliv1_payload(**self._data)
        self._location_builder = location_builder

    def to_json_api(self):
        attributes = dict(**self._data)
        if self._location_builder is not None:
            attributes.update(**self._location_builder._dliv2())

        # This is a file
        attributes.pop('sample_data', None)
        return {
            'type': 'dataset',
            'data': {'attributes': attributes}
        }

    def to_multipart_body(self):
        data = json.dumps(self.to_json_api())

        fields={
            'data': ('data', data, 'application/json'),
        }

        if self._data['sample_data']:
            fields['sample_data'] = (
                os.path.basename(self._data['sample_data'].name),
                self._data['sample_data'],
                # Mimetyppe is ignored
                'text/plain'
            )

        encoder = MultipartEncoder(fields=fields)

        # NOTE this really should be fixed on the catalogue as well.
        # the way to handle this and deliver an error message more
        # graceful than an arbitary socket closure is for the catalogue
        # to reject the requests with a 'Content-Length' header that's
        # too large.
        if encoder.len > (TEN_MEGABYTES - len(data)):
            raise ValueError(
                f'The sample_data file passed to the'
                ' DatasetBuilder is too large. Please '
                'keep files under 10 megabytes.'
            )

        return encoder

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A valid path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").
        :returns: DatasetBuilder (itself)
        :rtype: dli.client.components.dataset.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.components.dataset import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Structured",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                                data_preview_type='STATIC',
                                sample_data=open('path/to/my/file.csv')
                        )
                builder = builder.with_external_s3_storage(
                    bucket_name="external-s3-bucket-name",
                    aws_account_id=123456789,
                    prefix="/economic-data-package/my-test-dataset"
                )
                client = dli.connect()
                dataset = client.register_dataset(builder)
        """
        self._location_builder = DatasetLocationBuilder().with_external_s3_storage(
            bucket_name=bucket_name,
            aws_account_id=aws_account_id,
            prefix=prefix
        )
        self.payload.update(self._location_builder.build())
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :returns: DatasetBuilder (itself)
        :rtype: dli.client.components.dataset.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.components.dataset import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Structured",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                        )
                builder = builder.with_external_storage("external-storage-location")
                dataset = client.register_dataset(builder)
        """
        self._location_builder = DatasetLocationBuilder().with_external_storage(location)
        self.payload.update(self._location_builder.build())
        return self

    def build(self):
        return copy.copy(self.payload)

    @staticmethod
    def _validate_preview_type_static(preview_type):
        if preview_type != 'STATIC':
            raise ValueError(
                'Field: data_preview_type must be STATIC when '
                'sample_data is provided'
            )


class DatasetLocationBuilder:
    """
        A simple builder to specify dataset location.
    """

    def __init__(self):
        self.payload = {}

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A vaild path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").

        :returns: DatasetBuilder (itself)
        :rtype: dli.client.components.dataset.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.components.dataset import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_s3_storage(
                        bucket_name="external-s3-bucket-name",
                        aws_account_id=123456789,
                        prefix="/economic-data-package/my-test-dataset"
                    )
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "S3",
            "owner": {
                "awsAccountId": str(aws_account_id)
            },
            "bucket": bucket_name,
            "prefix": prefix
        }
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :returns: DatasetBuilder (itself)
        :rtype: dli.client.components.dataset.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.components.dataset import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "Other",
            "source": location
        }
        return self

    def _dliv2(self):
        """
        The old API used a polymorphic location type.
        The new API uses keys. While more simple we
        have to handle both cases.
        """
        data = {}
        location = None if 'location' not in self.payload else self.payload['location']
        if location and location['type'] == 'S3':
            data['location'] = {
                's3': copy.copy(self.payload['location'])
            }

            if 'owner' in data['location']['s3']:
                data['location']['s3']['owner']['aws_account_id'] = (
                    # Is camel cased in old API but not new api...
                    data['location']['s3']['owner'].pop('awsAccountId')
                )
        else:
            data['location'] = {
                'other': copy.copy(self.payload['location'])
            }

        return data

    def build(self):
        return copy.copy(self.payload)


def ensure_count_is_valid(count):
    count = int(count)
    if count <= 0:
        raise ValueError("`count` should be a positive integer")


# utilities above


class Dataset(BaseComponent):

    def get_s3_access_keys_for_dataset(self, *dataset_ids):
        """
        Retrieve S3 access keys for the specified account to access the
        specified dataset(s).

        :param list dataset_ids: One ore more dataset ids to get access to.
        :returns: A namedtuple containing the AWS keys and session token.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                s3_access_keys = client.get_s3_access_keys_for_dataset(dataset_id1, dataset_id2)
                # print(s3_access_keys)
                # access_key(access_key_id='39D19A440AFE452B9', secret_access_key='F426A93CDECE45C9BFF8F4F19DA5CB81', session_token='C0CC405803F244CA99999')

        """


        payload = {"datasetIds": list(dataset_ids)}

        response = self.session.post(
            dataset_urls.access_keys, json=payload
        )

        response_dict = response.json()
        obj = type(response_dict['class'][0], (_AttributeAdapter,), to_snake_case_keys(response_dict["properties"]))()
        return obj

    def _get_dataset_id_by_name(self, package_id=None,
                                dataset_name=None,
                                package_name=None):
        if dataset_name and package_name:
            params = {"name": package_name}

            p = self.packages(search_term=[f"name={package_name}"])
            if len(p) == 1:
                package_id = p[package_name].id
            else:
                raise CatalogueEntityNotFoundException("No such package")

        ds = self.datasets(search_term=[f"package_id={package_id}",
                                        f"name={dataset_name}"], warn=False)
        try:
            return next(iter(ds.values())).id
        except StopIteration as s:
            raise Exception(f"No such dataset `{dataset_name}`")

    def register_dataset(self, builder: DatasetBuilder):
        """
        Submit a request to create a new dataset under a specified package in the Data Catalogue.

        :param dli.client.components.dataset.DatasetBuilder builder: An instance of DatasetBuilder. This builder object sets sensible defaults and exposes
                                                           helper methods on how to configure its storage options.

        :returns: A newly created Dataset.
        :rtype: Dataset

        - **Sample**

        .. code-block:: python

                # Please refer to builder docs for examples on
                # how to create an instance of DatasetBuilder.
                client = dli.connect()
                dataset = client.register_dataset(builder)
        """
        mp_encoder = builder.to_multipart_body()

        result = self.session.post(
            dataset_urls.v2_index,
            data=mp_encoder,
            headers={
                'Content-Type': mp_encoder.content_type,
                'Content-Length': str(mp_encoder.len),
            },
        ).json()

        return self._DatasetFactory._from_v2_response(result)

    def edit_dataset(
        self,
        dataset_id,
        location_builder=None,
        **kwargs
    ):
        """
        Updates information on a dataset, returning the updated instance.
        If keyword argument is not specified field keeps its old value.
        Optional enum and text type fields can be unset by passing ``None``.

        :param str dataset_id: Id of the dataset being updated.
        :param DatasetLocationBuilder location_builder: Optional. An instance of DatasetLocationBuilder. This builder object exposes
                                                                            helper methods to configure dataset storage options.

        Keyword arguments:

        :keyword str name: Optional. A descriptive name of a dataset. It should be unique within the package.
        :keyword str description: Optional. A short description of a package.
        :keyword str content_type: Optional. A way for the data steward to classify the type of data in the dataset
                                (e.g. Structured or Unstructured).
        :keyword str data_format: Optional. The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `TSV`, XML`, `Other`.
        :keyword str publishing_frequency: Optional. The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
        :keyword list[str] keywords: Optional. User-defined list of keywords that can be used to find this dataset through
                                    the search interface.
        :keyword str naming_convention: Optional. Key for how to read the dataset name.
        :keyword str documentation: Optional. Documentation about this dataset in markdown format.
        :keyword list[str] taxonomy: Optional. A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
        :keyword str load_type: Optional. Whether each datafile in this dataset should be considered as a full version
                                of a dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`
        :keyword str data_preview_type: Optional enum of NONE STATIC LIVE. 'NONE' means there is no datapreview. STATIC means
                            to use the file provided. LIVE means to use the latest datafile.
        :keyword file sample_data: Optional file. File like object. Used in conjuntion with data_preview_type. Max size 10Megabytes.
        :keyword bool details_visible_to_lite_users: Optional. Default value "False".
        :keyword bool sql_enabled: Optional. Default value "False".


        :returns: Updated Dataset.
        :rtype: Dataset

        - **Sample**

        .. code-block:: python

                # e.g. update dataset description
                client = dli.connect()
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    description="Updated my dataset description"
                )

                # update dataset location. Please note that this is only allowed if the dataset has no datafiles registered.
                builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    location_builder=builder
                )

                # update dataset taxonomy
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    taxonomy=["Credit", "CDS"]
                )

        """
        builder = DatasetBuilder(
            location_builder=location_builder,
            **kwargs
        )

        mp_encoder = builder.to_multipart_body()
        result = self.session.patch(
            dataset_urls.v2_by_id.format(id=dataset_id), data=mp_encoder,
            headers={
                'Content-Type': mp_encoder.content_type,
                'Content-Length': str(mp_encoder.len),
            }
        ).json()

        return self._DatasetFactory._from_v2_response(result)

    def delete_dataset(self, dataset_id):
        """
        Marks a particular dataset (and all its datafiles) as deleted.
        This dataset will no longer be accessible by consumers.

        :param str dataset_id: The id of the dataset to be deleted.

        :returns: None

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                client.delete_dataset(dataset_id)

        """
        self.session.delete(
            dataset_urls.v2_by_id.format(id=dataset_id)
        )

    def get_datafiles(self, dataset_id, name_contains=None,
                      as_of_date_start=None, as_of_date_end=None,
                      page=1, count=100):
        """
        Returns a list of all datafiles registered under a dataset.

        :param str dataset_id: The id of the dataset.
        :param str name_contains: Optional. Look up only those datafiles for the dataset where name contains this string.
        :param str as_of_date_start: Optional. Datafiles having data_as_of date greater than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param str as_of_date_end: Optional. Datafiles having data_as_of date less than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param int page: Optional. The page to start retrieving from and ascending per internal paging. Defaults to 1.
        :param int count: Optional. The count of datafiles to be returned per internal paging. Defaults to 100.

        :returns: List of all datafiles registered under the dataset.
        :rtype: List[Datafile]

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                datafiles = client.get_datafiles(
                    dataset_id,
                    name_contains='My Test Data',
                    as_of_date_start='2018-10-11',
                    as_of_date_end='2018-10-15',
                    count=10
                )
        """

        ensure_count_is_valid(count)

        params = {
            'name': name_contains,
            'as_of_date_start': as_of_date_start,
            'as_of_date_end': as_of_date_end,
            'page_size': count,
            'page': page,
        }

        response_dict = self.session.get(
            dataset_urls.datafiles.format(id=dataset_id), params=params
        ).json()
        obj = [type('datafile', (_AttributeAdapter,), to_snake_case_keys(x["properties"]))()
                for x in response_dict['entities'] if x['rel'] == 'datafile']
        return obj

    def get_latest_datafile(self, dataset_id):
        """
        Fetches datafile metadata of latest datafile in the dataset.

        :param str dataset_id: The id of the dataset.

        :returns: The datafile.
        :rtype: Datafile

        - **Sample**

        .. code-block:: python

                client = dli.connect()
                datafile = client.get_latest_datafile(dataset_id)
        """
        response_dict = self.session.get(
                dataset_urls.latest_datafile.format(id=dataset_id)).json()

        obj = type(response_dict['class'][0], (_AttributeAdapter,), to_snake_case_keys(response_dict["properties"]))()
        return obj




#
#  Copyright (C) 2021 IHS Markit.
#  All Rights Reserved
#

# the below is ported directly from commons - we can't use commons in the SDK as its unpublished (authorisation)
# https://gitlab.ihsmarkit.com/data-lake/python-commons/-/blob/master/datalake/commons/exceptions/view_exceptions.py
#
# we could add some build process for the published SDK, but its already complicated enough
# Never the less, these classes are done for us, so we want to use them
from typing import Optional

try:
    from flask import g, request, Request
    request_opt: Optional[Request] = request
except ImportError:
    g = None
    request_opt = None

try:
    from ddtrace.helpers import get_correlation_ids
except ImportError:
    get_correlation_ids = lambda: (None, None)


class BaseHttpException(Exception):
    description: Optional[str] = None
    details: Optional[str] = None
    title = 'Unknown error'
    status_code = 500

    def __init__(self, description=None, details=None, title=None):
        Exception.__init__(self)
        if title:
            self.title = title

        if description:
            self.description = description

        if isinstance(details, dict):
            self.details = [{k: self._flatten_message(v)} for k, v in details.items()]
        elif isinstance(details, list):
            self.details = str.join(', ', details)
        elif isinstance(details, str):
            self.details = details

    def to_dict(self):
        trace_id, span_id = get_correlation_ids()

        return_value = {
            'status': self.status_code,
            'title': self.title,
            'description': self.description,
            'details': self.details,
            'trace_id': trace_id,
            'span_id': span_id,
        }

        if request_opt:
            return_value['request_id'] = getattr(g, 'request_id', None)

        return dict((k, v) for k, v in return_value.items() if v is not None)

    def _flatten_message(self, message):
        if isinstance(message, list):
            return str.join(', ', message)
        elif isinstance(message, dict):
            for key, value in list(message.items()):
                message[key] = self._flatten_message(value)

        return message


INTERNAL_SERVER_ERROR_MESSAGE = (
    'Something went wrong. Please contact the Data Lake technical team by '
    'raising a request with datalake-support@ihsmarkit.com.'
)


class BadRequest(BaseHttpException):
    status_code = 400
    description = 'Request was incorrect.'
    title = 'Bad request'


class Unauthorized(BadRequest):
    status_code = 401
    description = 'Access to the resource requires authentication.'
    title = 'Unauthorized'


class Forbidden(BadRequest):
    status_code = 403
    description = 'There are restrictions on this operation. Please contact our Support team (Datalake-Support-Dataonboarding@ihsmarkit.com) to discuss your permissions.'
    title = 'Forbidden'


class NotFound(BadRequest):
    status_code = 404
    description = 'Entity was not found.'
    title = 'Not found'


class Conflict(BadRequest):
    status_code = 409
    description = 'There is a conflict.'
    title = 'Conflict'


class UnprocessableEntity(BadRequest):
    status_code = 422
    description = 'Request body is not valid, see details.'
    title = 'Unprocessable entity'


class InternalServerError(BaseHttpException):
    status_code = 500
    description = INTERNAL_SERVER_ERROR_MESSAGE
    title = 'Internal server error'


class NotImplemented(BaseHttpException):
    status_code = 501
    description = 'Functionality required to fulfil request is not supported.'
    title = 'Not implemented'



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import textwrap
from collections import OrderedDict
from typing import Dict

import dli.models
from dli.models.dataset_model import DatasetModel

from platform_services_lib.lib.adapters.pagination_adapters import Paginator
from platform_services_lib.lib.services.dlc_attributes_dict import AttributesDict
from platform_services_lib.lib.context.urls import package_urls


class PackageModel(AttributesDict):
    """
    Represents a package
    """

    @classmethod
    def _from_v2_response_unsheathed(cls, response_json, page_size=None):
        """

        :param response_json:
        :param page_size: Unused value, added to keep the same interface as
        datasetModel's _from_v2_response_unsheathed.
        :return:
        """

        return cls({
            'data': response_json
        })

    def __init__(self, json):
        # We have to declare the attribute `documentation` because it is
        # referenced in the code of this class. This means that there will
        # be an `documentation` attribute even when there is zero
        # `documentation` attribute in the Catalogue response JSON.
        self.documentation = None
        self.__shape = None

        # Maintain compatibility between V1 (which had package_id) and V2
        # (which only has ID).
        if 'package_id' not in json['data']['attributes']:
            json['data']['attributes']['package_id'] = json['data']['id']

        super().__init__(id=json['data']['id'], **json['data']['attributes'])
        self._paginator = Paginator(
            package_urls.v2_package_datasets.format(id=self.id),
            self._client._DatasetFactory,
            self._client._DatasetFactory._from_v2_response_unsheathed
        )

    @property
    def shape(self):
        """
        :returns: The number of datasets within this package
        :rtype: Int
        """
        if self.__shape is None:
            self.__shape = len(self.datasets())
        return self.__shape

    def datasets(self) -> Dict[str, DatasetModel]:
        """
        :returns: Dictionary of datasets shortcodes mapping to the model of the respective dataset, for each in the package.
        :rtype: OrderedDict[id: str, dli.models.dataset_model.DatasetModel]
        """
        return OrderedDict([
            (v.short_code, v) for v in self._paginator
        ])

    def contents(self):
        """Print each of the datasets in this package.

          :Example:

            .. code:: python

                >>> p1 = client.packages.get('example')
                >>> p1.contents()

                DATASET "Paint Colours Palettes" [PARQUET]
                >> Shortcode: PaintColours
                >> Available Date Range: 2020-05-13 to 2020-05-13
                >> ID: 804002d8-2466-4303-a470-3cb73f6def77
                >> Published: Monthly by Dulux Paints
                >> Accessible: True

                This dataset lists the in vogue colour palettes

                --------------------------------------------------------------------------------
                DATASET "Furniture Inventory" [CSV]
                >> Shortcode: FurnitureInventory
                >> Available Date Range: 2012-01-13 to 2020-05-13
                >> ID: 804002d8-2466-4303-a470-3cb73f6aaab7
                >> Published: Daily by Furniture 4 You
                >> Accessible: True

                This dataset lists the currently in-stock furniture available to be purchase
                from the wholesaler for decorators and designers

                --------------------------------------------------------------------------------
        """
        for p in self.datasets().items():
            print(str(p[1]))

    def metadata(self):
        """
        Once you have selected a package, you can print the metadata
        (the available fields and values). These are the attributes and their values for this package as listed on this page

        :example:

            .. code-block:: python

                # Get all packages with a name containing the text 'example package'. Filtering
                # is done on the server side.
                >>> packages = client.packages('example package')
                # Get metadata of the 'example package' package.
                >>> packages['example package'].metadata()

        :return: Prints the metadata (attributes) keys and values.
        """
        dli.models.print_model_metadata(self)

    def __str__(self):
        separator = "-"*80
        split_description = "\n".join(textwrap.wrap(self.description, 80))
        split_keywords = "\n".join(self.keywords or [])

        # documentation is not guaranteed to be available
        split_documentation = 'No documentation available.'
        if self.documentation is not None:
            split_documentation = "\n".join(textwrap.wrap(self.documentation, 80))

        # When the fields are re-named in the JSON, it breaks our mapping. Use
        #   return str(self.__dict__)
        # to get the new field names then update the below.
        return f"\nPACKAGE \"{self.name}\" " \
               f"(Contains: {self.shape} datasets)\n" \
               f">> ID: {self.id} \n" \
               f">> Accessible: {self.has_access}\n" \
               f"\n" \
               f"{split_description}\n" \
               f"Documentation: {split_documentation}\n\n" \
               f"Keywords:\n{split_keywords}\n" \
               f"{separator}"

    def __repr__(self):
        return f'<Package name={self.name}>'
